__version__ = "0.5.2"

# from .builtins import *


import numpy as np

from .dataset import TrainingDataset
from .loglikelihood import LogLikelihood
from .minimize import Minimizer


class Cache:
    """A cache where the keys are numpy arrays

    Used to cache the results of the model computations, which may be
    computationnaly expensive. The cached values are indexed by the parameter
    vectors.

    This is not a generic class. Used in the models and the Likelohood, where
    you have states which you want to preserve, that are encoded by a vector of
    parameters.

    Why not use a hash ? Because I found that numpy.array_equal is in fact
    faster than hash(str(p)). If someone points me to a faster solution, I'll
    adopt it.

    """
    def __init__(self, max_size=2):
        self.max_size = max_size
        self._cache = []

    def __len__(self):
        return len(self._cache)

    def insert(self, p, blob):
        """insert a blob in cache, with p as a key

        Parameters:
        -----------
        p: (ndarray)
          parameter vector used as a key.
        blob: stuff
          the object to cache

        .. note:: this method does not verify that the key is not already in the
                  cache. If this is the case, then, the older blob will be
                  masked and the older (key,blob) entry will eventually be
                  pushed out of the cache. Note that if there is a value indexed
                  by the same key, then you might have a bug somewhere, because
                  it means that your function returns different values for the
                  same input parameter vector.
        """
        if len(self._cache) == self.max_size:
            self._cache.pop()
        self._cache.insert(0, (p, blob))

    def get(self, p):
        """retrieve a blob indexed by the parameter vector p from the cache

        if no match, return None.

        Parameters
        ----------
        p: (ndarray)
          parameter vector used as a key

        """
        for c in self._cache:
            key, blob = c
            if np.array_equal(p, key):
                return blob
        return None

    def keys(self):
        """return the list of keys in the cache
        """
        ret = [c[0] for c in self._cache]
        return ret



import logging
import pickle

import matplotlib as mpl
import numpy as np
import numpy.ma as ma
import pylab as pl
import pandas

from lemaitre import bandpasses
from nacl.dataset import TrainingDataset
from nacl.loglikelihood import LogLikelihood
from nacl.minimize import Minimizer
from nacl.models import salt2
from nacl.models.salt2 import constraints
from nacl.models.salt2.regularizations import NaClSplineRegularization
from nacl.models.salt2.variancemodels import SimpleErrorSnake, LocalErrorSnake

from saunerie.robuststat import mad



class FitResults:
    pass


class TrainSALT2Like:

    def __init__(self, tds, variance_model=None, mu_reg=1., mu_cons=1e-10):
        """
        Parameters:
        tds : nacl.dataset.TrainingDataset 
            Your training dataset
        variance_model : str or int or float
            if an error model needs to be fitted default is None
        mu_reg : float
            intensity of the regularization default is 1.
        mu_cons : float
            intensity of the penality of the constraints, default is 1e-10
        """
        self.tds = tds
        if self.tds.spec_data is not None:
            self._clean_dataset(tds)
        self.model = salt2.get_model(tds)
        if variance_model is not None:
            if variance_model == 'simple_snake': 
                variance_model = salt2.get_simple_snake_error(self.model)
                
            elif variance_model == 'local_snake':
                variance_model = salt2.get_local_snake_error(self.model)
                
            elif variance_model == 'sn_lambda_snake':
                variance_model = salt2.get_sn_lambda_snake_error(self.model)
            
            elif variance_model == 'sn_local_snake':
                variance_model = salt2.get_sn_local_snake_error(self.model)
            else:
                raise ValueError("Input error_model not recognised, try simple_snake, local_snake, sn_lambda_snake or sn_local_snake")
        self.variance_model = variance_model
        self.log = []
        self.mu_reg=1.
        self.mu_cons=1e-10

        # recalibrate the spectra
        if tds.spec_data is not None:
            ll = LogLikelihood(self.model)
            self.v_init = self.model(ll.pars)
            self.spec_recal_init = self._recalib_spectra(self.tds, self.v_init)
            # ll.pars['SpectrumRecalibration'].full[3::4] = np.log(r)

    def fit(self, fix=None, p_init=None,
            p_init_blocks=None,
            fit_variance_model=False,
            force_spec_recalibration=False,
            max_iter=100):
        """
        """
        reg = salt2.get_regularization_prior(self.model, pars=self.model.init_pars(), mu=self.mu_reg, order=1, check=True)

        cons = salt2.get_constraint_prior(self.model, linear=True,
                                      mu=self.mu_cons, Mb=-19.5,
                                      check=True)
        if fit_variance_model:
            ll = LogLikelihood(self.model, variance_model=self.variance_model)
            reg = salt2.get_regularization_prior(self.model, pars=ll.pars, mu=self.mu_reg, order=1, check=True)
            ll = LogLikelihood(self.model, variance_model=self.variance_model, reg=[reg], cons=[cons])
        else:
            ll = LogLikelihood(self.model, reg=[reg], cons=[cons])

        if force_spec_recalibration:
            ll.pars['SpectrumRecalibration'].full[3::4] = np.log(self.spec_recal_init)

        if fix:
            for block_name in fix:
                ll.pars[block_name].fix()

        if p_init:
            for block_name in p_init._struct.slices:
                if block_name not in ll.pars._struct.slices:
                    continue
                ll.pars[block_name].full[:] = p_init[block_name].full[:]

        if p_init_blocks:
            for block_name in p_init_blocks:
                if block_name not in ll.pars._struct.slices:
                    continue
                ll.pars[block_name].full[:] = p_init_blocks[block_name]

        # fit
        minz = Minimizer(ll)
        p = minz.minimize_lm(ll.pars.free, max_iter=max_iter)
        
        t = np.array(minz.get_log()['time'])
        logging.info(f'fitting took {t[-1] - t[0]} seconds')
        
        res = FitResults()
        res.pars = ll.pars
        res.ll = ll
        res.minz = minz
        res.v = self.model(res.ll.pars, jac=0)
        res.v_var = None
        if self.variance_model is not None and fit_variance_model:
            res.v_var = self.variance_model(res.ll.pars, jac=0)

        return res

    def train_salt2_model(self, save=False, path=None):
        """Full model training
        """
        full_pars_blocks = ['M0', 'M1', 'CL', 'X0', 'X1', 'c', 'tmax']
        spec_recal = False
        if self.tds.spec_data is not None:
            full_pars_blocks.append('SpectrumRecalibration')
            spec_recal = True
        # initialization: PETS sncosmo/salt2 fits
        f0 = FitResults()
        f0.ll = LogLikelihood(self.model)
        f0.pars = f0.ll.pars
        f0.v = self.model(f0.pars)
        f0.var_v = None
        self.log.append(f0)

        # initialization : fit the light curves and spectra
        # no error model
        f1 = self.fit(fix=['M0', 'M1', 'CL'],
                      fit_variance_model=False,
                      force_spec_recalibration=spec_recal,
                      max_iter=10)
        self.log.append(f1)

        # fit error model
        f2 = self.fit(fix=full_pars_blocks,
                      fit_variance_model=True,
                      p_init=f1.ll.pars,
                      max_iter=20)
        self.log.append(f2)

        # refit all pars,
        # with an error model
        f3 = self.fit(fix=None,
                      fit_variance_model=True,
                      p_init=f2.ll.pars,
                      max_iter=50)
        self.log.append(f3)

        # train
        #f4 = self.fit(variance_model=True,
        #              p_init=f3.ll.pars)
        #self.log.append(f4)

        if save and path is not None:
            sn_data = pandas.DataFrame(data=self.tds.sn_data.nt)
            columns_rename = {'x0':'x0_init', 'x1':'x1_init', 'c':'c_init', 'tmax':'tmax_init'}
            sn_data = sn_data.rename(columns = columns_rename)
            sn_data = sn_data.assign(x0_nacl = f3.pars['X0'].full[self.tds.sn_data.sn_index])
            sn_data = sn_data.assign(x1_nacl = f3.pars['X1'].full[self.tds.sn_data.sn_index])
            sn_data = sn_data.assign(c_nacl = f3.pars['c'].full[self.tds.sn_data.sn_index])
            sn_data = sn_data.assign(tmax_nacl = f3.pars['tmax'].full[self.tds.sn_data.sn_index])
            sn_data.to_parquet(path + '_params' + '.parquet')
            
            cov_mat = f3.minz.get_cov_matrix()[0]
            cov_reduced = cov_mat[:4*len(self.tds.sn_data), :4*len(self.tds.sn_data)]
            ii = [self.tds.sn_data.sn_index + i*len(self.tds.sn_data) for i in range(4)]
            iii = np.hstack(ii)
            cov_reduced = cov_reduced[:,iii][iii]
            np.save(path + '_cov' + '.npy', cov_reduced)
            
    def fit_err_model_test(self):
        """light curve fitting plus error model
        """
        full_pars_blocks = ['M0', 'M1', 'CL', 'X0', 'X1', 'c', 'tmax']
        if self.tds.spec_data is not None:
            full_pars_blocks.append('SpectrumRecalibration')
        # initialization: PETS sncosmo/salt2 fits
        f0 = FitResults()
        f0.ll = LogLikelihood(self.model)
        f0.pars = f0.ll.pars
        f0.v = self.model(f0.pars)
        f0.var_v = None
        self.log.append(f0)

        # initialization : fit the light curves and spectra
        # no error model
        f1 = self.fit(fix=['M0', 'M1', 'CL'],
                      fit_variance_model=False,
                      force_spec_recalibration=True,
                      max_iter=10)
        self.log.append(f1)

        # fit an error model, all other pars fixed
        f2 = self.fit(fix=full_pars_blocks,
                      fit_variance_model=True,
                      p_init=f1.ll.pars,
                      p_init_blocks={'gamma': -5.},
                      max_iter=10)
        self.log.append(f2)

        # refit the light curves and spectra,
        # with an error model
        f3 = self.fit(fix=['M0', 'M1', 'CL'],
                      fit_variance_model=True,
                      p_init=f2.ll.pars,
                      max_iter=10)
        self.log.append(f3)

        # train
        #f4 = self.fit(variance_model=True,
        #              p_init=f3.ll.pars)
        #self.log.append(f4)
        
    def _clean_dataset(self, tds, phase_range=(-20., +50.)):
        """
        """
        # clean ZTF19adcecwu
        # TODO: protocol to clean manually identified datapoints
        # idx = (tds.lc_data.sn == 'ZTF19adcecwu') & (tds.lc_data.mjd > 58840) & (tds.lc_data.flux<10000.)
        # logging.info(f'removing {idx.sum()} outliers identified on ZTF19adcecwu g- and r- DR2 lightcurves')
        # tds.lc_data.valid[idx] = 0

        # phase range (photometric data)
        tmax = np.zeros(len(tds.sn_data))
        tmax[tds.sn_data.sn_index] = tds.sn_data.tmax
        phot_tmax = tmax[tds.lc_data.sn_index]
        phase = (tds.lc_data.mjd - phot_tmax) / (1. + tds.lc_data.z)

        idx = (phase<phase_range[0]) | (phase>phase_range[1])
        logging.info(f'removing {idx.sum()} photometric points outside phase range')
        tds.lc_data.valid[idx] = 0

        # phase range (spectra)
        spec_tmax = tmax[tds.spec_data.sn_index]
        phase = (tds.spec_data.mjd - spec_tmax) / (1. + tds.spec_data.z)
        idx = (phase<phase_range[0]) | (phase>phase_range[1])
        logging.info(f'removing {idx.sum()} spectroscopic points outside phase range')
        tds.spec_data.valid[idx] = 0

        # points of the edge of the wavelength basis
        # no: this version does not what we want to do
        # i_basis_max = tds.spec_data.i_basis.max()
        # idx = (tds.spec_data.i_basis < 3) | (tds.spec_data.i_basis >= (i_basis_max-3))
        #
        # we need to determine, for each spectrum, the i_basis_min and i_basis_max
        b_bins = np.arange(-0.5, tds.spec_data.i_basis.max() + 1.5, 1)
        s_bins = np.arange(-0.5, tds.spec_data.spec_index.max() + 1.5, 1)
        h, _, _ = np.histogram2d(tds.spec_data.i_basis, tds.spec_data.spec_index, bins=(b_bins, s_bins))
        bb = np.arange(0, tds.spec_data.i_basis.max() + 1, 1)
        ss = np.arange(0, tds.spec_data.spec_index.max() + 1, 1)
        bb, ss = np.meshgrid(bb, ss)
        u = bb * h.T
        mb = ma.masked_array(u, mask=u==0.)
        b_min, b_max = np.array(np.min(mb, axis=1)), np.array(np.max(mb, axis=1))
        to_kill  = np.array(b_min).astype(int)[tds.spec_data.spec_index] == tds.spec_data.i_basis
        to_kill |= np.array(b_max).astype(int)[tds.spec_data.spec_index] == tds.spec_data.i_basis

        logging.info(f'removing {to_kill.sum()} spectroscopic points outside wavelength range')
        tds.spec_data.valid[to_kill] = 0

        # remove the spectral points with an uncertainty larger than 1.E6

        # clean all SNe below 0.02
        #idx = (tds.lc_data.z < 0.01)
        #logging.info(f'removing {idx.sum()} very low redshift SNe')
        #tds.lc_data.valid[idx] = 0

        tds.compress()
    def _recalib_spectra(self, tds, v):
        """rough recalibration of the spectra
        """
        vv = v[len(tds.lc_data):]
        ii = np.arange(len(vv))
        idx = np.abs(tds.spec_data.fluxerr / tds.spec_data.flux) < 1.
        zero_fluxerr = np.abs(tds.spec_data.fluxerr) <= 0.
        logging.info(f'ignoring {zero_fluxerr.sum()} points with buggy flux errs')
        idx &= ~zero_fluxerr
        neg_flux = tds.spec_data.flux <= 0.
        idx &= ~neg_flux

#        fig, axes = pl.subplots(nrows=3, ncols=1, figsize=(8, 10),
#                                sharex=True)
#        axes[0].errorbar(ii[idx], tds.spec_data.flux[idx],
#                         yerr=tds.spec_data.fluxerr[idx], ls='', marker='.')
#        axes[0].plot(ii[idx], vv[idx], marker='.', color='r')
        ii = vv[idx] == 0.
        r = np.abs(tds.spec_data.flux[idx][~ii] / vv[idx][~ii])
        r_err = tds.spec_data.fluxerr[idx][~ii] / np.abs(vv[idx][~ii])
#        axes[1].plot(ii[idx], r[idx],
#                     ls='', marker='.')

        w = 1. / r_err**2
        n = np.bincount(tds.spec_data.spec_index[idx][~ii], weights=(w*r))
        d = np.bincount(tds.spec_data.spec_index[idx][~ii], weights=w)
        recal = n / d

        v_recal = vv * recal[tds.spec_data.spec_index]

        #axes[2].errorbar(ii[idx], tds.spec_data.flux[idx],
        #                 yerr=tds.spec_data.fluxerr[idx], ls='', marker='.')
#        axes[2].plot(ii[idx], tds.spec_data.flux[idx],
#                     ls='', marker='.')
#        axes[2].plot(ii[idx], v_recal[idx], ls='', marker='.', color='r')

        return recal

    def _get_models_to_plot(self, numfit, pars):
        """
        """
        fits = []

        if numfit is not None:
            try:
                iter(numfit)
                fits.extend(self.log[i] for i in numfit)
            except:
                fits.append(self.log[numfit])

        if pars is not None:
            try:
                iter(pars)
                for p in pars:
                    ff = FitResults()
                    ff.pars = pars
                    ff.v = self.model(pars)
                    if self.variance_model:
                        ff.var_v = self.variance_model(pars)
                    fits.append(ff)
            except:
                ff = FitResults()
                ff.pars = pars
                ff.v = self.model(pars)
                if self.variance_model:
                    ff.var_v = self.variance_model(pars)
                fits.append(ff)

        return fits


    def plot_lc(self, sn, numfit=None, pars=None,
                phase=False, norm=1.,
                plot_variance=False):
        """plot light curve data + models

        Parameters
        ----------
        sn : (int | str)
          the supernova to plot
        numfit : int, optional
          which fit to plot (drawn from logs)
        pars : FitParameters, optional
          plot an alternate fit, not from self.log
        phase : bool, default False
          whether to plot the the light curve as a function of the phase
        plot_variance : bool, default=False
          whether to plot the error model, if available

        Returns
        -------
        None
        """
        sel = self.tds.lc_data.sn == sn
        bands = np.unique(self.tds.lc_data.band[sel]).tolist()
        bands.sort(key=_get_band_order)
        nrows = len(bands)

        sn_idx = self.tds.sn_data.sn == sn
        tmax = float(self.tds.sn_data.tmax[sn_idx])
        z = float(self.tds.sn_data.z[sn_idx])
        sn_index = int(self.tds.sn_data.sn_index[sn_idx])

        # clone the dataset
        t = clone(self.tds, sn)
        to_plot = self._get_models_to_plot(numfit, pars)
        m = salt2.get_model(t)
        if self.variance_model is not None:
            vm = self.variance_model.__class__(m)
        else:
            vm = None
        plotters = [ModelPlotter(m, sn_index, tp.ll.pars, variance_model=vm) \
                    for tp in to_plot]

        # plots
        fig, axes = pl.subplots(nrows=nrows, ncols=1,
                                figsize=(8,12),
                                sharex=True)
        for i,b in enumerate(bands):
            idx = sel & (self.tds.lc_data.band == b)
            wl = self.tds.lc_data.wavelength[idx].mean()
            color = pl.cm.jet(int((wl-2000)/(11000.-2000.) * 256))

            if phase:
                xx = (self.tds.lc_data.mjd[idx] - tmax) / (1. + z)
                axes[i].axvline(0., ls=':')
            else:
                xx = self.tds.lc_data.mjd[idx]
                axes[i].axvline(tmax, ls=':')

            axes[i].errorbar(xx, self.tds.lc_data.flux[idx],
                             yerr=self.tds.lc_data.fluxerr[idx],
                             ls='', marker='.', color=color)

            for p in plotters:
                p.plot(b, ax=axes[i],
                       color=color,
                       phase=phase,
                       ylabel=b)
            if phase:
                axes[-1].set_xlabel('mjd [days]')
            else:
                axes[-1].set_xlabel('phase [restframe days]')
            pl.subplots_adjust(hspace=0.05)
            fig.suptitle(f'{sn} @ z={z:4.3}')

    def plot_spec(self, spec, numfit=None, pars=None, alpha=0.5):
        """
        """
        sel = self.tds.spec_data.spec == spec

        fig, axes = pl.subplots(nrows=1, ncols=1,
                                figsize=(8,8),
                                sharex=True)

        axes.errorbar(self.tds.spec_data.wavelength[sel], self.tds.spec_data.flux[sel],
                      yerr=self.tds.spec_data.fluxerr[sel],
                      ls='', marker='.', color='blue')

        to_plot = self._get_models_to_plot(numfit, pars)
        for tp in to_plot:
            vv = tp.v[len(self.tds.lc_data):][sel]
            axes.plot(self.tds.spec_data.wavelength[sel], vv, 'r+:')
            if tp.v_var is not None:
                vv_var = tp.v_var[len(self.tds.lc_data):][sel]
                vmin = vv - np.sqrt(vv_var)
                vmax = vv + np.sqrt(vv_var)
                axes.fill_between(self.tds.spec_data.wavelength[sel],
                                  vmin, vmax,
                                  color='r', alpha=alpha)

        axes.set_ylabel('flux')
        axes.set_xlabel('mjd')

        sn = np.unique(self.tds.spec_data.sn[sel])
        assert len(sn) == 1
        sn = sn[0]
        z = np.unique(self.tds.spec_data.z[sel])
        assert len(z) == 1
        z = z[0]
        axes.set_title(f'{sn} @ z={z:4.3}')
        pl.subplots_adjust(hspace=0.05)

    def cut_lc_outliers(self, pars, nsig=10):
        """
        """
        v = self.model(pars)
        if self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)
        else:
            var_v = None

        vv = v[:len(self.tds.lc_data)]
        if var_v is not None:
            var_vv = var_v[:len(self.tds.lc_data)]
        else:
            var_vv = None

        bands = np.unique(self.tds.lc_data.band)
        nbands = len(bands)

        for i,band in enumerate(bands):
            sel = self.tds.lc_data.band == band
            var = self.tds.lc_data.fluxerr**2
            if var_vv is not None:
                var += var_vv
            w = 1. / var
            res = vv - self.tds.lc_data.flux
            wres = np.sqrt(w) * res

            cut = (np.abs(wres) > nsig)
            cc = cut[sel]
            logging.info(f'cutting {cc.sum()}/{len(cc)}={100*cc.sum()/len(cc):4.3}% outliers in {band}')
            self.tds.lc_data.valid[sel] = (~cut)[sel]

            jj = (~cut)[sel]
            print(jj.sum(), self.tds.lc_data.valid[sel].sum(), len(self.tds.lc_data.valid[sel]))


    def plot_lc_training_residuals(self, pars, v=None, var_v=None,
                                   phases=False,
                                   weighted_residuals=False,
                                   hexbin=False):
        """global fit residuals
        """
        if not v:
            v = self.model(pars)
        if not var_v and self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)

        vv = v[:len(self.tds.lc_data)]
        if var_v is not None:
            var_vv = var_v[:len(tds.lc_data)]
        else:
            var_vv = None

        tmax = pars['tmax'].full[self.tds.lc_data.sn_index]
        z = self.tds.lc_data.z

        bands = np.unique(self.tds.lc_data.band)
        nbands = len(bands)
        fig, axes = pl.subplots(figsize=(6,12),
                                nrows=nbands, ncols=2,
                                sharex=0, sharey=1,
                                gridspec_kw={'width_ratios': [3,1]})
        for i, band in enumerate(bands):
            sel = self.tds.lc_data.band == band
            wl = self.tds.lc_data.wavelength[sel].mean()
            color = pl.cm.jet(int((wl-2000)/(11000.-2000.) * 256))
            if phases:
                x = (self.tds.lc_data.mjd - tmax) / (1. + z)
            else:
                x = np.arange(len(sel))

            var = self.tds.lc_data.fluxerr**2
            if var_vv is not None:
                var += var_vv
            w = 1. / var
            res = vv - self.tds.lc_data.flux
            wres = np.sqrt(w) * res
            y = wres if weighted_residuals else res
            if hexbin:
                axes[i,0].hexbin(x[sel], y[sel],
                                 gridsize=(100, 500),
                                 extent=(-30, 50, -100, 100),
                                 mincnt=1)
            else:
                axes[i,0].plot(x[sel], y[sel],
                               # yerr=tds.lc_data.fluxerr[sel],
                               ls='', marker='.', color=color)
            axes[i,1].hist(y[sel], bins=500,
                           color=color,
                           density=True,
                           # orientation='vertical')
                           )
            print(band, mad(y[sel]))

            # if var_vv is not None:
            #    ii = np.argsort(x[sel])
            # axes[i].fill_between(x[sel][ii], -np.sqrt(var_vv[sel][ii]), np.sqrt(var_vv[sel][ii]),
            #                      alpha=0.25, color=color)
            axes[i,0].set_ylabel(band)
        axes[-1,0].set_xlabel('phases [days]' if phases else 'index')
        pl.subplots_adjust(hspace=0.01, wspace=0.01)

        return v, var_v


    def plot_spec_training_residuals(self, pars, v=None, var_v=None):
        """global fit residuals
        """
        if not v:
            v = self.model(pars)
        if not var_v and self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)

        vv = v[len(self.tds.lc_data):]
        if var_v is not None:
            var_vv = var_v[len(self.tds.lc_data):]
        else:
            var_vv = None

        sel = self.tds.spec_data.valid

        tmax = pars['tmax'].full[self.tds.spec_data.sn_index]
        z = self.tds.spec_data.z
        phases = (self.tds.spec_data.mjd - tmax) / (1. + z)

        fig, axes = pl.subplots(nrows=1, ncols=1)
        # color = pl.cm.jet(int(phases/(50+25) * 256))
        x = self.tds.spec_data.wavelength
        ii = np.argsort(x)
        axes.errorbar(x[sel], vv[sel] - self.tds.spec_data.flux[sel],
                      yerr=self.tds.spec_data.fluxerr[sel],
                      ls='', marker=',', color='gray')
        #        if var_vv is not None:
        #            axes[i].fill_between(x[sel][ii], -np.sqrt(var_vv[sel][ii]), np.sqrt(var_vv[sel][ii]),
        #                                 alpha=0.25, color='gray')
        axes.set_xlabel('wavelength [$\AA$]')
        pl.subplots_adjust(hspace=0.05)

        return v, var_v


    def plot_lightcurves(self, pars, v=None, var_v=None, phases=False):
        """global fit residuals
        """
        if not v:
            v = self.model(pars)
        if not var_v and self.variance_model is not None and 'gamma' in pars._struct.slices:
            var_v = self.variance_model(pars)

        vv = v[:len(self.tds.lc_data)]
        if var_v is not None:
            var_vv = var_v[:len(self.tds.lc_data)]
        else:
            var_vv = None

        tmax = pars['tmax'].full[self.tds.lc_data.sn_index]
        z = self.tds.lc_data.z

        bands = np.unique(self.tds.lc_data.band)
        nbands = len(bands)
        fig, axes = pl.subplots(nrows=nbands, ncols=1, sharex=phases)
        for i, band in enumerate(bands):
            sel = self.tds.lc_data.band == band
            wl = self.tds.lc_data.wavelength[sel].mean()
            color = pl.cm.jet(int((wl-2000)/(9000.-2000.) * 256))
            if phases:
                x = (self.tds.lc_data.mjd - tmax) / (1. + z)
            else:
                x = self.tds.lc_data.mjd
            ii = np.argsort(x[sel])

            axes[i].errorbar(x[sel][ii], self.tds.lc_data.flux[sel][ii],
                             yerr=self.tds.lc_data.fluxerr[sel][ii],
                             ls='', marker=',', color=color)
            axes[i].plot(x[sel][ii], vv[sel][ii],
                         ls='', marker='+', color=color)

            if var_vv is not None:
                axes[i].fill_between(x[sel][ii], vv[sel]-np.sqrt(var_vv[sel][ii]), vv[sel]+np.sqrt(var_vv[sel][ii]),
                                     alpha=0.5, color=color)
            axes[i].set_ylabel(band)
        axes[-1].set_xlabel('phases [days]' if phases else 'index')
        pl.subplots_adjust(hspace=0.05)

        return v, var_v

    def save(self, fn):
        with open(fn, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, fn):
        with open(fn, 'rb') as f:
            return pickle.load(f)

    def plot(self, res):
        """
        """
        t = TrainingChi2(self.tds, res.ll.pars, res.ll.model)
        d = t.get_sorted_list(sort_by='sn')

        # plot the 5 worst SNe
        for nm in d[:5].sn:
            plot_lc(nm, self.tds, res.ll.pars, norm=1, phase=1)

        for spec in [222, 333, 444, 555, 1226, 1500, 1700]:
            plot_spec(spec, self.tds, res.v)


_band_ordering = {'u': 0, 'U': 0, 'g': 1, 'B': 2, 'r': 3, 'r2': 3,
                  'R': 4, 'i': 5, 'i2': 5, 'I': 6, 'z': 7, 'Y': 8}

def _get_band_order(band):
    b = band.split('::')[-1]
    return _band_ordering.get(b, 100)


def order_bands(bands):
    l = bands.sort(key=_get_band_order)
    return l


def clone(tds, sn, phase_range=(-25, 51., 1.)):
    """
    """
    # sn metadata
    sn_data = tds.sn_data.nt[tds.sn_data.sn == sn]

    # lightcurve data
    if tds.lc_data is not None:
        lc_data = tds.lc_data.nt[tds.lc_data.sn == sn]

        tmax = float(sn_data.tmax)
        z = float(sn_data.z)
        phase = np.arange(*phase_range)
        mjd = phase * (1.+z) + tmax
        N = len(mjd)
        l = []
        for b in np.unique(lc_data.band):
            d = lc_data[lc_data.band == b]
            Z = np.full(len(mjd), d[0])
            Z['mjd'] = mjd
            Z['flux'] = 0.
            Z['fluxerr'] = 0.
            Z['valid'] = 1
            l.append(Z)
        lc_data = np.rec.array(np.hstack(l))
    else:
        lc_data = None

    if tds.spec_data is not None:
        spec_data = tds.spec_data.nt[tds.spec_data.sn == sn]
    else:
        spec_data = None
    if tds.spectrophotometric_data is not None:
        spectrophotometric_data = tds.spectrophotometric_data.nt[tds.spectrophotometric_data.sn == sn]
    else:
          spectrophotometric_data = None


    tds = TrainingDataset(sn_data, lc_data=lc_data,
                          spec_data=None,
                          spectrophotometric_data=None,
                          filterlib=tds.filterlib)

    return tds


class ModelPlotter:
    """A utility class to plot a smooth continuous (oversampled) photometric model
    """

    def __init__(self, model, sn_index, init_pars, variance_model=None):
        """Constructor - evaluate the model and error model
        """
        self.model = model
        self.tds = model.training_dataset
        self.init_pars = init_pars
        self.variance_model = variance_model
        self._init_local_pars(init_pars, sn_index)

        # evaluate model and variance model
        self.v = self.model(self.pars)
        if variance_model is not None:
            self.v_var = self.variance_model(self.pars)
        else:
            self.v_var = None

    def _init_local_pars(self, init_pars, sn_index):
        """initialize small parameter vector from many-sne one
        """
        if self.variance_model is not None:
            ll = LogLikelihood(self.model, variance_model=self.variance_model)
            self.pars = ll.pars.copy()
        else:
            self.pars = self.model.init_pars()

        for nm in ['M0', 'M1', 'CL']:
            self.pars[nm].full[:] = init_pars[nm].full[:]
        for nm in ['X0', 'X1', 'c', 'tmax']:
            self.pars[nm].full[0] = init_pars[nm].full[sn_index]
        if 'gamma' in init_pars._struct.slices:
            self.pars['gamma'].full[:] = init_pars['gamma'].full[:]
        if 'gamma_sn' in init_pars._struct.slices:
            self.pars['gamma_sn'].full[0] = init_pars['gamma_sn'].full[sn_index]
        if 'gamma_snake' in init_pars._struct.slices:
            self.pars['gamma_snake'].full[:] = init_pars['gamma_snake'].full[:]
        # else:
        #    self.pars['gamma'].full[:] = 0.

        return self.pars

    def plot(self, band, ax=None, **kwargs):
        """plotter - plot the model and error model if needed
        """
        if not ax:
            ax = pl.gca()

        phase = kwargs.get('phase', False)
        color = kwargs.get('color', 'b')
        ls = kwargs.get('ls', ':')
        label = kwargs.get('label', '')
        xlabel= kwargs.get('xlabel', None)
        ylabel= kwargs.get('ylabel', None)
        title = kwargs.get('title', None)
        alpha = kwargs.get('alpha', 0.25)
        legend = kwargs.get('legend', None)

        tmax = float(self.tds.sn_data.tmax[0])
        z = float(self.tds.sn_data.z[0])

        idx = self.tds.lc_data.band == band
        if phase:
            xx = (self.tds.lc_data.mjd - tmax) / (1. + z)
            ax.axvline(0., ls=':')
        else:
            xx = self.tds.lc_data.mjd
            ax.axvline(tmax, ls=':')

        ax.plot(xx[idx], self.v[idx], color=color, ls=ls, label=label)
        if self.variance_model is not None:
            vm = (self.v[idx] - np.sqrt(self.v_var[idx]))
            vp = (self.v[idx] + np.sqrt(self.v_var[idx]))
            ax.fill_between(xx[idx], vm, vp, alpha=0.5, color=color)

        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if title is not None:
            ax.set_title(title)
        if legend:
            ax.legend(loc=legend)



#    rVdVVr             tr_VdV            tr_VdVVdV
#    1                  1                 1             ok. brent.
#    1                  1                -1             cholesky failed.
#    1                 -1                 1             converged. wrong res.
#    1                 -1                -1             cholesky failed
#   -1                  1                 1             brent. wrong res.
#   -1                  1                -1             cholesky failed.
#   -1                 -1                 1             brent wrong res.
#   -1                 -1                -1             cholesky failed.




import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

from scipy import sparse
import numpy as np
import pylab as pl
import pandas
from nacl.lib.fitparameters import FitParameters
from nacl.dataset import TrainingDataset
from nacl.models.salt import SALT2Like
from nacl.models.constraints import SALT2LikeConstraints
from nacl.models.regularizations import NaClSplineRegularization
from nacl.minimize import LogLikelihood2, Minimizer
import helpers


class AxBTrainingDataset:
    """Simplistic training dataset for tests
    """
    def __init__(self, N=100, a=1., b=0., sigma_meas=0.1, sigma_unknown=1.):
        """constructor - initialize a recarray with simulated data
        """
        x = np.random.uniform(-1, 1, N)
        sigma_tot = np.sqrt(sigma_meas**2 + sigma_unknown**2)
        print(sigma_meas, sigma_unknown, sigma_tot)
        noise = np.random.normal(scale=sigma_tot, size=N)
        y = a * x + b + noise
        ey = np.full(N, sigma_meas)
        valid = np.full(N, 1)
        self.data = np.rec.fromarrays((x, y, ey, valid),
                                      names=('x', 'y', 'ey', 'valid'))

    def nb_meas(self, valid_only=False):
        if valid_only:
            return self.data['valid'].sum()
        return len(self.data)

    def get_all_fluxes(self):
        return self.data['y']

    def get_all_fluxerr(self):
        return self.data['ey']

    def get_valid(self):
        return self.data['valid']

    def plot(self, ax=None, model=None):
        if ax is None:
            fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(8,8))
        axes.errorbar(self.data['x'], self.data['y'],
                      yerr=self.data['ey'],
                      ls='', marker='.', color='k')
        if model is not None:
            a = model.pars['a'].full[0]
            b = model.pars['b'].full[0]
            pl.plot(self.data['x'], a*self.data['x']+b, 'r-')
        return axes

class Model:

    def __init__(self, training_dataset, var_class=None):
        self.training_dataset = training_dataset
        self.data = training_dataset.data
        if var_class is not None:
            self.variance_model = var_class(self)
        self.pars = self.init_pars()
        if self.variance_model is not None:
            self.variance_model.init_pars(self.pars)

    def get_struct(self):
        return [('a', 1), ('b', 1)]

    def init_pars(self):
        s = self.get_struct()
        if self.variance_model:
            s += self.variance_model.get_struct()
        fp = FitParameters(s)
        fp['a'] = 1.
        return fp

    def __call__(self, p, jac=False):
        a, b = self.pars['a'].full[0], self.pars['b'].full[0]
        v = a * self.data['x'] + b
        if not jac:
            return v

        model_vals = v

        N = len(self.data)
        i,j,v = [], [], []
        i.append(np.arange(N))
        j.append(np.full(N, self.pars['a'].indexof()))
        v.append(self.data['x'])

        i.append(np.arange(N))
        j.append(np.full(N, self.pars['b'].indexof()))
        v.append(np.full(N, 1.))

        i = np.hstack(i)
        j = np.hstack(j)
        v = np.hstack(v)
        idx = j >= 0
        n_free_pars = len(self.pars.free)
        J = sparse.coo_matrix((v[idx], (i[idx],j[idx])),
                              shape=(N,n_free_pars))

        return model_vals, J


class VarianceModel:

    def __init__(self, model):
        self.training_dataset = model.training_dataset
        self.data = model.data

    def get_struct(self):
        return [('sigma', 1)]

    def init_pars(self, pars):
        self.pars = pars
        self.pars['sigma'].full[:] = 0.5

    def __call__(self, jac=False, model_flux=None, model_jac=None):
        """
        """
        N = len(self.data)
        sigma = self.pars['sigma'].full[0]
        v = np.full(N, sigma**2)
        if not jac:
            return v

        i = np.arange(N)
        j = np.full(N, self.pars['sigma'].indexof())
        vv = np.full(N, 2. * sigma)
        idx = j >= 0
        n_free_pars = len(self.pars.free)
        J = sparse.coo_matrix((vv[idx], (i[idx], j[idx])),
                               shape=(N, n_free_pars))
        return v, J

    def noise(self, N):
        v = self(jac=False)
        return np.random.normal(scale=np.sqrt(v))

def plot_chi2(chi2):
    """
    """
    chi2.debug = True
    pars = chi2.model.pars
    x = np.linspace(0.2, 5., 100)
    full_chi2 = []
    main_chi2 = []
    log_det_v = []
    for xx in x:
        pars['sigma'].full[0] = xx
        full_chi2.append(chi2(pars.free))
        main_chi2.append(chi2.chi2_debug)
        log_det_v.append(chi2.log_det_v_debug)
    pl.plot(x, full_chi2, 'k.-')
    pl.plot(x, main_chi2, 'r:')
    pl.plot(x, log_det_v, 'g:')


class AixBiCDTrainingDataset:
    """Simplistic training dataset for tests
    """
    def __init__(self, nlc=10, npts=100,
                 a_i=None, b_i=None,
                 sigma_meas=0.1, sigma_unknown=1.):
        """constructor - initialize a recarray with simulated data
        """
        self.nlc = nlc
        self.a_i = a_i
        if a_i is None:
            self.a_i = np.random.uniform(- 5.,  5., nlc)
        self.b_i = b_i
        if b_i is None:
            self.b_i = np.random.uniform(-10., 10., nlc)
        self.c = - self.b_i.sum()
        self.npts = npts
        self.sigma_meas = sigma_meas
        self.sigma_unknown = sigma_unknown

        i_, x_, y_, ey_, valid_ = [], [], [], [], []
        for i in range(nlc):
            x = np.random.uniform(-1, 1, npts)
            sigma_tot = np.sqrt(sigma_meas**2 + sigma_unknown**2)
            noise = np.random.normal(scale=sigma_tot, size=npts)
            y = self.a_i[i] * x + self.b_i[i] + self.c + noise
            ey = np.full(npts, sigma_meas)
            i_.append(np.full(npts, i))
            x_.append(x)
            y_.append(y)
            ey_.append(ey)
            valid_.append(np.full(npts, 1))

        self.data = np.rec.fromarrays(
            (np.hstack(i_), np.hstack(x_), np.hstack(y_),
             np.hstack(ey_), np.hstack(valid_)),
                names=('i', 'x', 'y', 'ey', 'valid'))

    def nb_meas(self, valid_only=False):
        if valid_only:
            return self.data['valid'].sum()
        return len(self.data)

    def get_all_fluxes(self):
        return self.data['y']

    def get_all_fluxerr(self):
        return self.data['ey']

    def get_valid(self):
        return self.data['valid']

    def plot(self, ax=None, model=None):
        if ax is None:
            fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(8,8))
        axes.errorbar(self.data['x'], self.data['y'],
                      yerr=self.data['ey'],
                      ls='', marker='.', color='k')
        if model is not None:
            a = model.pars['a'].full[0]
            b = model.pars['b'].full[0]
            pl.plot(self.data['x'], a*self.data['x']+b, 'r-')
        return axes

class AiBiCDModel:

    def __init__(self, training_dataset, var_class=None):
        self.training_dataset = training_dataset
        self.data = training_dataset.data
        self.variance_model = None
        if var_class is not None:
            self.variance_model = var_class(self)
        self.pars = self.init_pars()
        if self.variance_model is not None:
            self.variance_model.init_pars(self.pars)

    def get_struct(self):
        nlc = self.training_dataset.nlc
        return [('a_i', nlc), ('b_i', nlc), ('c', 1) , ('d', 1)]

    def init_pars(self):
        s = self.get_struct()
        if self.variance_model:
            s += self.variance_model.get_struct()
        fp = FitParameters(s)
        fp['a_i'] = 1.
        return fp

    def __call__(self, p, jac=False):
        tds = self.training_dataset
        a_i = self.pars['a_i'].full[tds.data.i]
        b_i = self.pars['b_i'].full[tds.data.i]
        c = self.pars['c'].full[0]
        d = self.pars['d'].full[0]
        v = a_i * self.data['x'] + b_i + c  + d
        if not jac:
            return v

        model_vals = v

        N = len(self.data)
        i,j,v = [], [], []
        i.append(np.arange(N))
        j.append(np.full(N, self.pars['a_i'].indexof(tds.data.i)))
        v.append(self.data['x'])

        i.append(np.arange(N))
        j.append(np.full(N, self.pars['b_i'].indexof(tds.data.i)))
        v.append(np.full(N, 1.))

        i.append(np.arange(N))
        j.append(np.full(N, self.pars['c'].indexof()))
        v.append(np.full(N, 1))

        i.append(np.arange(N))
        j.append(np.full(N, self.pars['d'].indexof()))
        v.append(np.full(N, 1))

        i = np.hstack(i)
        j = np.hstack(j)
        v = np.hstack(v)
        idx = j >= 0
        n_free_pars = len(self.pars.free)
        J = sparse.coo_matrix((v[idx], (i[idx],j[idx])),
                              shape=(N,n_free_pars))

        return model_vals, J


class Constraint:

    def __init__(self, model, rhs, pars_struct=None):
        """
        """
        self.model = model
        self.rhs = rhs
        if pars_struct:
            self.pars_struct = pars_struct
        else:
            self.pars_struct = self.model.pars.copy()
            self.pars_struct.release()

    def __call__(self, p=None, deriv=False):
        pass


class LinearConstraint(Constraint):

    def __init__(self, model, h_matrix, rhs, pars_struct=None):
        """
        """
        super().__init__(model, rhs, pars_struct)
        self.h_matrix = h_matrix
        self.rhs = rhs
        # if pars_struct:
        #     self.pars_struct = pars_struct
        # else:
        #     self.pars_struct = self.model.pars.copy()
        #     self.pars_struct.release()

    def __call__(self, p=None, deriv=False):
        """
        """
        cons = self.h_matrix @ self.model.pars.full - self.rhs
        cons = float(cons)
        if not deriv:
            return cons
        return cons, self.h_matrix, None


class ModelBCD2Constraint(Constraint):

    def __init__(self, model, rhs, pars_struct=None):
        super().__init__(model, rhs, pars_struct)

    def __call__(self, p=None, deriv=False):
        """
        """
        pars = self.model.pars
        cons = pars['b_i'].full**2 + pars['c'].full**2 + pars['d'].full**2
        cons = cons.sum()
        cons -= self.rhs

        if not deriv:
            return cons

        # if deriv
        npars_full = len(self.model.pars.full)
        j = np.hstack((self.pars_struct['b_i'].indexof(),
                       self.pars_struct['c'].indexof(),
                       self.pars_struct['d'].indexof()))
        v = np.hstack((2. * pars['b_i'].full,
                       2. * pars['c'].full,
                       2. * pars['d'].full))
        i = np.zeros(len(j))
        dcons = sparse.coo_matrix((v, (i,j)), shape=(1, npars_full))

        # second derivatives
        d2cons = sparse.coo_matrix((np.full(len(j), 2.), (j,j)),
                          shape=(npars_full, npars_full))

        return cons, dcons, d2cons


class ConstraintSet:
    """Combine a series of constraints to produce a (quadratic) penality,
    added to the Log Likelihood. Compute the gradient and the hessian
    of this penality if required.
    """
    def __init__(self, model, constraints):
        """constructor
        """
        self.model = model
        self.constraints = constraints

    def __call__(self, p=None, deriv=False):
        """evaluate the penality
        """
        if p is not None:
            self.model.pars.free = p

        npars = len(self.model.pars.full)

        pen = 0.
        # if no derivatives specified, return the sum of the quadratic
        # penalities associated with each constraint
        if not deriv:
            for cons in self.constraints:
                pen += cons(p=p, deriv=False)**2
            return float(pen)

        # otherwise, compute and return the gradient and hessian
        # along with the quadratic penality
        grad = sparse.coo_matrix(([], ([], [])), shape=(1,npars))
        hess = sparse.coo_matrix(([], ([],[])), shape=(npars,npars))
        for cons in self.constraints:
            c, dc, d2c = cons(p=p, deriv=True)
            pen  += c**2
            # TODO: return +grad and change the minimizer.
            # here, we return -grad, which is misleading
            grad += -2. * float(c) * dc
            hess +=  2. * dc.T.dot(dc)
            if d2c is not None:
                hess += 2. * c * d2c

        # fixed parameters ?
        idx = self.model.pars.indexof() >= 0
        pen = float(pen)
        grad = np.array(grad[:,idx].todense()).squeeze()
        hess = hess[:,idx][idx,:]

        return pen, grad, hess


def sum_bi(model, pars_struct=None):
    """
    """
    pars = model.pars
    npars = len(pars.full)

    i, j, v = [], [], []
    j = pars['b_i'].indexof()
    i = np.full(len(j), 0)
    v = np.full(len(j), 1.)
    mat = sparse.coo_matrix((v, (i,j)), shape=(1, npars)).tocsr()

    return LinearConstraint(model, mat, 0., pars_struct)


def test_aibicd_d_fixed_linear_constraint():
    tds = AixBiCDTrainingDataset(sigma_unknown=2.1)
    model = AiBiCDModel(tds, var_class=VarianceModel)
    model.pars['d'].fix()
    c = sum_bi(model)
    cons = ConstraintSet(model, [c])
    ll = LogLikelihood2(model, variance_model=model.variance_model, cons=[cons])
    minz = Minimizer(ll)
    ret = minz.minimize(model.pars.free)
    ret['model'] = model
    ret['cons'] = cons
    return ret


class AixBiCDConstraints:

    def __init__(self, model, mu=1.E10):
        self.model = model
        self.mu = mu
        self.pars_full = self.model.pars.copy()
        self.pars_full.release()
        self.rhs = np.array([0.])  # , 1000.])
        self.active = {}
        self.ncons = 1
        self.dh = {}
        self.dh[0] = self._init_dh_0()

    def _init_dh_0(self):
        """precompute the matrix of the first constraints
        """
        npars = len(self.model.pars.full)
        pars = self.pars_full

        # \sum b_i + c = 0
        i, j, v = [], [], []
        j = np.hstack((pars['b_i'].indexof(), pars['c'].indexof()))
        i = np.full(len(j), 0)
        v = np.full(len(j), 1.)
        return sparse.coo_matrix((v, (i,j)), shape=(1, npars)).tocsr()

    def eval_0(self, deriv=False):
        """
        """
        h = float(self.dh[0] @ self.model.pars.full[:])
        if not deriv:
            return h
        return h, self.dh[0], None

    def eval_1(self, deriv=False):
        """
        """
        pars = self.pars_full
        npars = len(pars.full)

        h = (pars['b_i'].full**2).sum() + \
             pars['c'].full[0]**2
             # pars['d'].full[0]**2

        if not deriv:
            return h

        # first derivatives
        i, j, v = [], [], []
        j.append(pars['b_i'].indexof())
        j.append(pars['c'].indexof())
        # j.append(pars['d'].indexof())
        j = np.hstack(j)
        i = np.zeros(len(j))
        v.append(2. * pars['b_i'].full)
        v.append(2. * pars['c'].full)
        # v.append(2. * pars['d'].full)
        v = np.hstack(v)
        dh = sparse.coo_matrix((v, (i, j)),
                                shape=(1, npars)).tocsr()
        # second derivatives
        i, j, v = [], [], []
        i = np.hstack((pars['b_i'].indexof(),
                       pars['c'].indexof(),
                       # pars['d'].indexof())
                       ))
        j = i
        v = np.full(len(i), 2.)
        d2h = sparse.coo_matrix((v, (i, j)),
                                shape=(npars, npars))

        return h, dh, None

    def check_active(self):
        """if some parameters are all fixed, the related constraints
        should probably be deactivated (?)
        """
        pass

    def __call__(self, p, deriv=False):
        """
        """
        self.model.pars.free = p
        self.pars_full.full[:] = self.model.pars.full[:]
        n_pars_full = len(self.model.pars.full)
        h = np.zeros(self.ncons)
        if not deriv:
            for i,f in enumerate([self.eval_0]):  # self.eval_1
                h[i] = f(deriv=False)
            cons = h - self.rhs
            return self.mu * (h**2).sum()

        dh  = sparse.coo_matrix(([], ([], [])),
                                shape=(self.ncons, n_pars_full)).tocsr()
        d2h = []

        # call all the constraints
        # in fact, we could precalculate the linear ones
        # and only call the non-linear ones.
        # but ok. no difference.
        # d2_available = False
        for i, f in enumerate([self.eval_0]):  # , self.eval_1]):
            h_, dh_, d2h_ = f(deriv=True)
            h[i] += h_
            dh[i,:] += dh_
            d2h.append(d2h_)

        cons = h - self.rhs
        penality = cons.T @ cons

        idx = self.model.pars.indexof()>=0
        grad = -2. * dh.T.dot(cons)[idx]
        hess =  2. * dh.T.dot(dh)[idx,:][:,idx]
        for i,d2 in enumerate(d2h):
            if d2 is not None:
                hess += 2. * cons[i] * d2

        return self.mu * penality, self.mu * grad, self.mu * hess

    # def __call__(self, p, deriv=False):
    #     """
    #     """
    #     self.model.pars.free = p
    #     self.pars_full.full[:] = self.model.pars.full[:]

    #     cons = self.Jac @ self.model.pars.full
    #     cons -= self.rhs

    #     cons_t_cons = float(cons.T @ cons)
    #     if not deriv:
    #         return self.mu * cons_t_cons

    #     JJ = self.Jac
    #     grad = -2. * self.mu * JJ.T.dot(cons)
    #     hess =  2. * self.mu * JJ.T.dot(JJ)

    #     return self.mu * cons_t_cons, grad, hess

def plot_reco(rec_data, tds, par_num=0, title=''):

    fig, axes = pl.subplots(figsize=(8,8), nrows=2, ncols=3)

    delta = rec_data.a_i.apply(lambda x: x[par_num]) - tds.a_i[par_num]
    axes[0,0].hist(delta, bins=20)
    axes[0,0].axvline(ls=':', color='r')
    axes[0,0].set_xlabel('$\Delta a$')

    delta = rec_data.b_i.apply(lambda x: x[par_num]) - tds.b_i[par_num]
    axes[0,1].hist(delta, bins=20)
    axes[0,1].axvline(ls=':', color='r')
    axes[0,1].set_xlabel('$\Delta b$')

    delta = rec_data.c - tds.c
    axes[0,2].hist(delta, bins=20)
    axes[0,2].axvline(ls=':', color='r')
    axes[0,2].set_xlabel('$\Delta c$')

    delta = rec_data.sigma - tds.sigma_unknown
    axes[1,0].hist(delta, bins=20)
    axes[1,0].axvline(ls=':', color='r')
    axes[1,0].set_xlabel('$\Delta \sigma$')

    axes[1,1].hist(rec_data.cons, bins=20)
    axes[1,1].set_xlabel('constraints')

    axes[1,2].hist(rec_data.chi2/rec_data.ndof)
    axes[1,2].set_xlabel('$\chi^2 / ndof$')

    if title:
        fig.suptitle(title)


def main_aixbicd(nlc=10, n_realizations=1, npts=100, plot=False,
                 a_i=None, b_i=None,
                 sigma_meas=0.25, sigma_unknown=1.2):
    """
    """
    ret = []
    if a_i is None:
        a_i = np.random.uniform(-10., 10., nlc)
    if b_i is None:
        b_i = np.random.uniform(-10., 10., nlc)
    for i in range(n_realizations):
        tds = AixBiCDTrainingDataset(nlc=nlc, npts=npts,
                                     a_i=a_i, b_i=b_i,
                                     sigma_meas=sigma_meas,
                                     sigma_unknown=sigma_unknown)
        model = AiBiCDModel(tds, var_class=VarianceModel)
        model.pars['d'].fix()
        cons = AixBiCDConstraints(model, mu=1E10)
        ll = LogLikelihood2(model,
                            variance_model=model.variance_model,
                            cons=[cons])
        minz = Minimizer(ll)
        n_free_pars = len(model.pars.free)
        model.pars.full[:] = np.random.uniform(-100., 100, n_free_pars)
        model.pars['sigma'].full[:] = 10.
        res = minz.minimize(model.pars.free)
        ret.append(((np.array(list(map(float, model.pars['a_i'].full))),
                   np.array(list(map(float, model.pars['b_i'].full))),
                   np.array([model.pars['c'].full[0]]),
                   np.array([model.pars['sigma'].full[0]]),
                   float(cons(model.pars.free)),
                   float(res['chi2']),
                   float(res['ndof']),
                   res['status'])))

    cols = ['a_i', 'b_i', 'c', 'sigma']
    # for k, slc in model.pars._struct.slices.items():
    #     n = slc.stop-slc.start
    #     if n == 1:
    #         cols.append(k)
    #     else:
    #         for i in range(n):
    #             cols.append(f'{k}_{i}')
    cols += ['cons', 'chi2', 'ndof', 'status']
    ret = pandas.DataFrame.from_records(ret, columns=cols,
                                        coerce_float=True)

    if plot:
        plot_reco(ret, tds,
                  title=f'AixBiCDModel [{n_realizations} realizations]')

    return ret, tds


def main_axb(N_realizations=1, N=100, plot=False):
    """
    """
    res = []
    for i in range(N_realizations):
        data = AxBTrainingDataset(N=100)
        model = Model(data, var_class=VarianceModel)
        ll = LogLikelihood2(model, variance_model=model.variance_model)
        minz = Minimizer(ll)
        model.pars.full[0] = 2.
        model.pars.full[1] = 2.
        model.pars.full[2] = 4.
        # model.pars.fix()
        # model.pars['sigma'].release()
        # model.pars['sigma'].fix()
        # s = minz.minimize(model.pars.free)
        # model.pars.release()
        # model.pars.fix()
        # model.pars['sigma'].release()
        s = minz.minimize(model.pars.free)
        pp = s['pars']
        res.append((pp['a'].full[0], pp['b'].full[0], pp['sigma'].full[0]))

    res = np.rec.fromrecords(res, names=['a', 'b', 'sigma'])

    if plot:
        fig, axes = pl.subplots(nrows=2, ncols=2, figsize=(8,8))
        axes[0,0].hist(res['a'], bins=20)
        axes[0,0].set_xlabel('a')
        axes[0,0].set_title('a')
        axes[0,0].axvline(1, ls=':', color='r')

        axes[0,1].hist(res['b'], bins=20)
        axes[0,1].set_xlabel('b')
        axes[0,1].set_title('b')
        axes[0,1].axvline(0, ls=':', color='r')

        axes[1,0].hist(res['sigma'], bins=20)
        axes[1,0].set_xlabel('$\sigma$')
        axes[1,0].set_title('$\sigma$')
        axes[1,0].axvline(1, ls=':', color='r')

    return res, s


if __name__ == '__main__':
    res, s = main_axb(1)





#!/usr/bin/env python3

import numpy as np
import pandas
import pylab as pl
from lemaitre import bandpasses
from nacl import TrainingDataset
from nacl.models.salt2 import SALT2Like


if __name__ == '__main__':

    fl = bandpasses.get_filterlib()

    sn_data = pandas.read_parquet('sn_data.parquet',
                                  columns=['z', 'x0', 'x1', 'c', 't0', 'ra', 'dec'],
                                  ).rename(columns={'t0': 'tmax'})
    #sn_data = sn_data_orig[[]]
    #sn_data.rename(columns={'t0': 'tmax'}, inplace=True)
    sn_data['sn'] = sn_data.index
    sn_data['mwebv'] = 0.
    sn_data['survey'] = ''
    sn_data['valid'] = 1

    lc_data = pandas.read_parquet('lc_data.parquet',
                                  columns=['z', 'mjd', 'band'])
    #lc_data = lc_data_orig[['snname', 'z', 'mjd', 'band']]
    #lc_data.rename(columns={'snname': 'sn'}, inplace=True)
    lc_data['sn'] = lc_data.index
    lc_data['exptime'] = 0.
    lc_data['valid'] = 1
    lc_data['flux'] = 0.
    lc_data['fluxerr'] = 0.
    lc_data['magsys'] = 0.
    lc_data['lc'] = lc_data['sn'].astype(str) + '-' + lc_data['band']
    lc_data['seeing'] = 0.
    lc_data['zp'] = 0.
    lc_data['mag_sky'] = 0.
    lc_data['x'] = 0.
    lc_data['y'] = 0.
    # lc_data['ccd'] = 0.
    # lc_data['amp'] = 0.
    lc_data['sensor_id'] = 0

    print('training dataset...')
    tds = TrainingDataset(sne=sn_data.to_records(),
                          lc_data=lc_data.to_records(),
                          filterlib=fl)
    tds.plot_sample()



import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

from scipy import sparse
import numpy as np
import pylab as pl
import pandas
from nacl.lib.fitparameters import FitParameters
from nacl.dataset import TrainingDataset
from nacl.models.salt import SALT2Like
from nacl.models.constraints import SALT2LikeConstraints
from nacl.models.regularizations import NaClSplineRegularization
from nacl.minimize import LogLikelihood2, Minimizer
import helpers


class AxBTrainingDataset:
    """Simplistic training dataset for tests
    """
    def __init__(self, N=100, a=1., b=0., sigma_meas=0.1, sigma_unknown=1.):
        """constructor - initialize a recarray with simulated data
        """
        x = np.random.uniform(-1, 1, N)
        sigma_tot = np.sqrt(sigma_meas**2 + sigma_unknown**2)
        print(sigma_meas, sigma_unknown, sigma_tot)
        noise = np.random.normal(scale=sigma_tot, size=N)
        y = a * x + b + noise
        ey = np.full(N, sigma_meas)
        valid = np.full(N, 1)
        self.data = np.rec.fromarrays((x, y, ey, valid),
                                      names=('x', 'y', 'ey', 'valid'))

    def nb_meas(self, valid_only=False):
        if valid_only:
            return self.data['valid'].sum()
        return len(self.data)

    def get_all_fluxes(self):
        return self.data['y']

    def get_all_fluxerr(self):
        return self.data['ey']

    def get_valid(self):
        return self.data['valid']

    def plot(self, ax=None, model=None):
        if ax is None:
            fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(8,8))
        axes.errorbar(self.data['x'], self.data['y'],
                      yerr=self.data['ey'],
                      ls='', marker='.', color='k')
        if model is not None:
            a = model.pars['a'].full[0]
            b = model.pars['b'].full[0]
            pl.plot(self.data['x'], a*self.data['x']+b, 'r-')
        return axes


class Model:

    def __init__(self, training_dataset, var_class=None):
        self.training_dataset = training_dataset
        self.data = training_dataset.data
        if var_class is not None:
            self.variance_model = var_class(self)
        self.pars = self.init_pars()
        if self.variance_model is not None:
            self.variance_model.init_pars(self.pars)

    def get_struct(self):
        return [('a', 1), ('b', 1)]

    def init_pars(self):
        s = self.get_struct()
        if self.variance_model:
            s += self.variance_model.get_struct()
        fp = FitParameters(s)
        fp['a'] = 1.
        return fp

    def __call__(self, p, jac=False):
        a, b = self.pars['a'].full[0], self.pars['b'].full[0]
        v = a * self.data['x'] + b
        if not jac:
            return v

        model_vals = v

        N = len(self.data)
        i,j,v = [], [], []
        i.append(np.arange(N))
        j.append(np.full(N, self.pars['a'].indexof()))
        v.append(self.data['x'])

        i.append(np.arange(N))
        j.append(np.full(N, self.pars['b'].indexof()))
        v.append(np.full(N, 1.))

        i = np.hstack(i)
        j = np.hstack(j)
        v = np.hstack(v)
        idx = j >= 0
        n_free_pars = len(self.pars.free)
        J = sparse.coo_matrix((v[idx], (i[idx],j[idx])),
                              shape=(N,n_free_pars))

        return model_vals, J


class VarianceModel:

    def __init__(self, model):
        self.training_dataset = model.training_dataset
        self.data = model.data

    def get_struct(self):
        return [('sigma', 1)]

    def init_pars(self, pars):
        self.pars = pars
        self.pars['sigma'].full[:] = 0.5

    def __call__(self, jac=False, model_flux=None, model_jac=None):
        """
        """
        N = len(self.data)
        sigma = self.pars['sigma'].full[0]
        v = np.full(N, sigma**2)
        if not jac:
            return v

        i = np.arange(N)
        j = np.full(N, self.pars['sigma'].indexof())
        vv = np.full(N, 2. * sigma)
        idx = j >= 0
        n_free_pars = len(self.pars.free)
        J = sparse.coo_matrix((vv[idx], (i[idx], j[idx])),
                               shape=(N, n_free_pars))
        return v, J

    def noise(self, N):
        v = self(jac=False)
        return np.random.normal(scale=np.sqrt(v))


def main_axb(N_realizations=1, N=100, plot=False):
    """
    """
    res = []
    for i in range(N_realizations):
        data = AxBTrainingDataset(N=100)
        model = Model(data, var_class=VarianceModel)
        ll = LogLikelihood2(model, variance_model=model.variance_model)
        minz = Minimizer(ll)
        model.pars.full[0] = 2.
        model.pars.full[1] = 2.
        model.pars.full[2] = 4.
        # model.pars.fix()
        # model.pars['sigma'].release()
        # model.pars['sigma'].fix()
        # s = minz.minimize(model.pars.free)
        # model.pars.release()
        # model.pars.fix()
        # model.pars['sigma'].release()
        s = minz.minimize(model.pars.free)
        pp = s['pars']
        res.append((pp['a'].full[0], pp['b'].full[0], pp['sigma'].full[0]))

    res = np.rec.fromrecords(res, names=['a', 'b', 'sigma'])

    if plot:
        fig, axes = pl.subplots(nrows=2, ncols=2, figsize=(8,8))
        axes[0,0].hist(res['a'], bins=20)
        axes[0,0].set_xlabel('a')
        axes[0,0].set_title('a')
        axes[0,0].axvline(1, ls=':', color='r')

        axes[0,1].hist(res['b'], bins=20)
        axes[0,1].set_xlabel('b')
        axes[0,1].set_title('b')
        axes[0,1].axvline(0, ls=':', color='r')

        axes[1,0].hist(res['sigma'], bins=20)
        axes[1,0].set_xlabel('$\sigma$')
        axes[1,0].set_title('$\sigma$')
        axes[1,0].axvline(1, ls=':', color='r')

    return res, s


if __name__ == '__main__':
    res, s = main_axb(1)






import numpy as np
from scipy.sparse import coo_matrix
from ....lib.fitparameters import FitParameters
from ....lib import bspline



class Model:
    """A 2D spline description of 2D data

    Encpsulating the spline evaluation in a model, like that
    allows to easily fix parameters, use the nacl minimizer,
    with regularization penalties.
    """
    def __init__(self, x0, x1, grid_x0=None, grid_x1=None, order=4):
        self.x0 = x0
        self.x1 = x1
        self.grid_x0 = grid_x0
        self.grid_x1 = grid_x1
        self.basis = bspline.BSpline2D(grid_x0, grid_x1, x_order=order, y_order=order)
        X0, X1 = np.meshgrid(x0, x1)
        self.J = self.basis.eval(X0.ravel(), X1.ravel())

    def get_struct(self):
        return [('theta', len(self.basis))]

    def init_pars(self, pars=None):
        if pars is None:
            pars = FitParameters(self.get_struct())
        n = len(pars['theta'].full)
        pars['theta'].full[:] = np.random.rand(n)
        return pars

    def __call__(self, pars, jac=False):
        """
        """
        vals = self.J @ pars['theta'].full
        if not jac:
            return vals

        j = pars.indexof(self.J.col)
        idx = j>=0

        N = self.J.shape[0]
        n_free_pars = len(pars.free)
        JJ = coo_matrix((self.J.data[idx], (self.J.row[idx], j[idx])), shape=(N,n_free_pars))
        return vals, JJ



"""Error snake examples

"""


import numpy as np
from scipy import sparse


class SimplePedestalErrorModel:
    """Error snake variance

    A simple model: just a constant pedestal error added to the data
    """
    def __init__(self, model):
        """constructor
        """
        self.model = model

    def get_struct(self):
        """structure of the error model specific parameters
        """
        return [("gamma", 1)]

    def init_pars(self, pars):
        pars['gamma'].full[:] = 1.

    def __call__(self, pars, jac=False):
        """
        """
        v, J = None, None
        if jac:
            v, J = self.model(pars, jac=True)
        else:
            v = self.model(pars, jac=False)
        N = len(v)
        g = pars['gamma'].full[0]

        var = np.full(N, g**2)

        if not jac:
            return var

        # n_pars = len(pars.full)
        n_free_pars = len(pars.free)
        dvar = np.full(N, 2. * g)
        i = np.arange(N)
        j = pars['gamma'].indexof(np.zeros(N).astype(int))
        idx = j>=0
        J = sparse.coo_matrix((dvar[idx], (i[idx], j[idx])),
                              shape=(N,n_free_pars))

        return var, J


class ConstantErrorSnake:
    """Simplistic Error snake

    This model is slightly more complicated: the variance depends on the model value:

    ..math::
        \sigma = \gamma \times f(\lambda,p)
    """
    def __init__(self, model):
        pass

    def get_struct(self):
        return [('gamma', 1)]

    def __call__(self, p, jac=False):
        pass


class VariableErrorSnake:
    """Realistic Error Snake

    This model is even more complicated: the variance depends on the model value
    and varies as a function of math:`$x$`

    """
    def __init__(self, model, basis=None):
        pass

    def get_struct(self):
        pass

    def __call__(self, p, jac=False):
        pass



"""SALT2 model: spectral part


TODO: remove color law evaluation from SpectrumEvalUnitFast
TODO: rename SpectrumEvalUnitFast

"""
import time
import numpy as np
import scipy

from ...sparseutils import CooMatrixBuff2


class SpectrumEvalUnitFast(object):
    r"""Evaluate the model for all SN spectra in the training dataset

    This class is one of the two type of "eval units". Given a chunk of the
    training dataset which corresponds the spectral observations, and given the
    :math:`(X_0, X_1, c, t_{max})` parameters of the SN, it computes the
    quantity:

    .. math::
         \frac{1}{1+z} \left[M_0\left(\frac{\lambda}{1+z}, \mathrm{p}\right) + X_1\ M_1\left(\frac{\lambda}{1+z},
         \mathrm{p}\right) \right]\ 10^{0.4\ c\ P(\frac{\lambda}{1+z})}\ s(\lambda_{rec})

    where
    .. math::
        M_{0|1}(\lambda, \mathrm{p}) = \sum_{k\ell} \theta_{k\ell} B_k(\mathrm{p}) B_l(\mathrm{\lambda})

    and
    .. math::
         \mathrm{p} = \frac{t - t_{max}}{1+z}

    and where :math:`R_s(\lambda)` is a polynomial correction which absorbs the
    wavelength-dependent large scale calibration errors affecting the spectrum.

    The evaluation reduces to a sparse matrix multiplication.

    """
    def __init__(self, tds, model, spec_recal_degree=3):
        """Constructor.

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.training_dataset = tds
        self.spec_data = tds.spec_data
        self.model = model

        # self.spec_index = self.data.spec_index
        # self.sn_index = self.data.sn_index
        # self.z = self.training_dataset.sn_data.z[self.sn_id]
        self.z = self.spec_data.z

        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)

        # restframe wavelengths
        self.restframe_wl = self.spec_data.wavelength/(1.+self.z)
        #        self.Jl = model.basis.bx.eval(self.restframe_wl).tocsr()

        # and we can connect directly to the global parameters
        #        self.M0 = model.pars['M0'].full
        #        self.M1 = model.pars['M1'].full
        #        self.cl_pars = model.pars['CL'].full

        # recalibration polynomial (one per spectrum, because we adapt to the spectrum wavelength range)
        #        self.recal_func = SpectrumRecalibrationPolynomials(self.training_dataset, self.model,
        #                                                           self.model.recalibration_degree)
        # self.recal_func = self.model.recal_func
        self.recal_func = SpectrumRecalibrationPolynomials(model, deg=spec_recal_degree)
        self.model.recal_func = self.recal_func

    def __call__(self, pars, jac=False, debug_mode=False):
        r"""Evaluate the model on all spectra

        Parameters
        ----------
        jac : bool
            if True compute and return the jacobian matrix
        debug_mode : bool
            if true, just return the model components.

        Returns
        -------
        val : numpy.array
            model evaluations.
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (if jac is true).
        """
        t0 = time.perf_counter()
        spec_data = self.training_dataset.spec_data
        sn_index = spec_data.sn_index
        basis = self.model.basis
        color_law = self.model.color_law

        # pars = self.model.pars
        M0 = pars['M0'].full
        M1 = pars['M1'].full
        cl_pars = pars['CL'].full

        # sn-related parameters
        x0, x1 = pars['X0'].full[sn_index], pars['X1'].full[sn_index]
        c, tmax = pars['c'].full[sn_index],  pars['tmax'].full[sn_index]

        # we need to re-evaluate the basis on the phases, since tmax changes
        restframe_phases = (spec_data.mjd-tmax)/(1.+self.z)
        jacobian = basis.eval(self.restframe_wl, restframe_phases + self.model.delta_phase).tocsr()

        # model components
        component_0 = jacobian.dot(M0)
        component_1 = jacobian.dot(M1)
        polynome_color_law, jacobian_color_law = \
            color_law(self.restframe_wl, cl_pars, jac=jac)
        color_law = np.power(10., 0.4*c*polynome_color_law)
        zz = 1. + self.z

        # recalibration polynomial
        # if we evaluate this, then recal_func must be instantiated
        # if self.recal_func is not None:
        assert self.recal_func is not None
        recal, jacobian_spec_rec = self.recal_func(pars, jac=jac)
        if jacobian_spec_rec is not None:
            jacobian_spec_rec = jacobian_spec_rec.tocoo()
            # don't know what to do with this
            # jacobian_spec_rec.data *= recal[jacobian_spec_rec.row]

        # recal = np.exp(recal)
        pca = (component_0 + x1 * component_1)
        model = pca * color_law * recal / zz

        if debug_mode:
            return component_0, component_1, color_law, recal, model

        self.model.val.append(model)
        if not jac:
            self.model.timing.append(time.perf_counter()-t0)
            return model

        jacobian = jacobian.tocoo()
        # X0 does not appear in the spectral part of the model
        # hence, dmdX0 = 0

        # dMdX1
        jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        jacobian_i.append(spec_data.row)
        jacobian_j.append(pars['X1'].indexof(sn_index))
        jacobian_val.append(component_1 * color_law * recal / zz)

        # dMdc
        jacobian_i.append(spec_data.row)
        jacobian_j.append(pars['c'].indexof(sn_index))
        jacobian_val.append(model * 0.4 * np.log(10.) * polynome_color_law)

        # dMdtmax
        # we can gain a little here, by not evaluating the gradient along the wavelength (ddlambda)
        _, deval_phase = self.model.basis.gradient(self.restframe_wl, restframe_phases + self.model.delta_phase)
        deval_phase = deval_phase.tocsr()
        jacobian_i.append(spec_data.row)
        jacobian_j.append(pars['tmax'].indexof(sn_index))
        jacobian_val.append(-1. * (deval_phase.dot(M0) + x1*deval_phase.dot(M1)) * color_law * recal / zz**2)

        # dmdtheta_0
        jacobian_i.append(spec_data.row[jacobian.row])
        jacobian_j.append(pars['M0'].indexof(jacobian.col))
        jacobian_val.append(jacobian.data * color_law[jacobian.row] * recal[jacobian.row] / zz[jacobian.row])

        # dmdtheta_1
        jacobian_i.append(spec_data.row[jacobian.row])
        jacobian_j.append(pars['M1'].indexof(jacobian.col))
        jacobian_val.append(x1[jacobian.row] * jacobian.data * color_law[jacobian.row] *
                            recal[jacobian.row] / zz[jacobian.row])

        # dMdcl (color law)
        jacobian_color_law = scipy.sparse.coo_matrix(jacobian_color_law)
        jacobian_i.append(spec_data.row[jacobian_color_law.row])
        jacobian_j.append(pars['CL'].indexof(jacobian_color_law.col))
        jacobian_val.append(c[jacobian_color_law.row] * 0.4 * np.log(10.) * jacobian_color_law.data *
                            model[jacobian_color_law.row])

        # dMdr (recalibration)
        if jacobian_spec_rec is not None:
            jacobian_i.append(spec_data.row[jacobian_spec_rec.row])
            jacobian_j.append(pars['SpectrumRecalibration'].indexof(jacobian_spec_rec.col))
            jacobian_val.append(jacobian_spec_rec.data * (pca*color_law)[jacobian_spec_rec.row]/zz[jacobian_spec_rec.row])

        if self.model.disable_cache:
            self.model.clear_cache()
        self.model.timing.append(time.perf_counter()-t0)

        return model


class CompressedSpectrumEvalUnit(object):
    """Predicts the components of the projection of the spectra on the model basis
    """
    def __init__(self, model, spec_recal_degree=3):
        """Constructor.

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.model = model
        self.training_dataset = model.training_dataset
        self.recal_func = SpectrumRecalibrationPolynomials(model, deg=spec_recal_degree)
        self.model.recal_func = self.recal_func

    def __call__(self, pars, jac=False):
        r"""Evaluate the model on all spectra

        Parameters
        ----------
        pars : `saltworks.FitParameters`
            the model parameters`
        jac : bool
            if True compute and return the jacobian matrix

        Returns
        -------
        val : numpy.array
            model evaluations.
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (if jac is true).
        """
        tds = self.training_dataset
        spec_index = tds.spec_db.spec_index
        mjd = tds.spec_db.mjd
        sn_index = tds.spec_db.sn_index
        z = tds.spec_db.z

        # restframe phases
        restframe_phase = (mjd - pars['tmax'].full[sn_index]) / (1. + z)
        phase_basis = self.model.basis.by
        J = phase_basis.eval(restframe_phase + self.model.delta_phase)

        # TODO: we use the wavelength field to store the spline basis number
        # this is not a great idea, I think. We should use wavelength to
        # store the spline function mean wavelength and add a new field to store
        # the basis function number
        i_basis = tds.spec_data.i_basis.astype(int)

        # predict the spectrum coefficients
        n_wl, n_ph = len(self.model.basis.bx), len(self.model.basis.by)
        M0 = pars['M0'].full.reshape(n_ph, n_wl)
        MM = J.dot(M0)
        V0 = MM[tds.spec_data.spec_index, i_basis]

        M1 = pars['M1'].full.reshape(n_ph, n_wl)
        MM = J.dot(M1)
        V1 = MM[tds.spec_data.spec_index, i_basis]

        # model parameters
        X1 = pars['X1'].full[tds.spec_data.sn_index]

        # recalibration parameters
        recal, Jr = self.recal_func(pars, jac=jac)
        recal = np.exp(recal)

        zz = 1. + tds.spec_data.z
        pca = V0 + X1 * V1
        model_val = pca * recal / zz


        if not jac:
            v = np.zeros(len(self.training_dataset))
            v[tds.spec_data.row] = self.model.norm * model_val
            return v

        # evaluate the derivatives
        N = len(tds)
        n_free_pars = len(pars.free)
        buff = CooMatrixBuff2((N, n_free_pars))


        # X0 does not appear in the spectral part model of the
        # dmdX0 = 0

        # dMdX1
        # J_i, J_j, J_val = [], [], []
        # J_i.append(tds.spec_data.row)
        # J_j.append(pars['X1'].indexof(tds.spec_data.sn_index))
        # J_val.append(V1 * recal / zz)
        buff.append(tds.spec_data.row,
                    pars['X1'].indexof(tds.spec_data.sn_index),
                    V1 * recal / zz)

        # no color law (absorbed by recal)
        # dMdc = 0

        # dtmax
        # we can gain a little here, by not evaluating the gradient along the
        # wavelength (ddlambda)
        dJ = phase_basis.deriv(restframe_phase + self.model.delta_phase)
        dV0 = dJ.dot(M0)[tds.spec_data.spec_index, i_basis]
        dV1 = dJ.dot(M1)[tds.spec_data.spec_index, i_basis]
        # dJ_phase = dJ_phase.tocsr()
        # J_i.append(tds.spec_data.row)
        # J_j.append(pars['tmax'].indexof(tds.spec_data.sn_index))
        # J_val.append(-1. * (dV0 + X1 * dV1) * recal / zz**2)
        buff.append(tds.spec_data.row,
                    pars['tmax'].indexof(tds.spec_data.sn_index),
                    -1. * (dV0 + X1 * dV1) * recal / zz**2)

        del dJ
        del dV0
        del dV1

        # dM0
        j = np.arange(n_ph)
        row = tds.spec_data.row.repeat(n_ph)
        tds_row = np.arange(len(tds.spec_data)).repeat(n_ph)
        spec_index = tds.spec_data.spec_index.repeat(n_ph)
        i_basis = tds.spec_data.i_basis.repeat(n_ph)
        l = np.tile(np.arange(n_ph), len(tds.spec_data))
        col = l * n_wl + i_basis # or maybe the other way around
        val = np.array(J.tocsr()[spec_index, l]).squeeze()

        # J_i.append(row)
        # J_j.append(pars['M0'].indexof(col))
        # J_val.append(val * recal[tds_row] / zz[tds_row])
        buff.append(row,
                    pars['M0'].indexof(col),
                    val * recal[tds_row] / zz[tds_row])

#        J_i.append(tds.spec_data.row[row])
#        J_j.append(pars['M0'].indexof(col))
#        J_val.append(val * recal[row] / zz[row])

        # dM1
        # J_i.append(row)
        # J_j.append(pars['M1'].indexof(col))
        # J_val.append(X1[tds_row] * val * recal[tds_row] / zz[tds_row])
        buff.append(row,
                    pars['M1'].indexof(col),
                    X1[tds_row] * val * recal[tds_row] / zz[tds_row])

#        J_i.append(tds.spec_data.row[row])
#        J_j.append(pars['M1'].indexof(col))
#        J_val.append(X1[row] * val * recal[row] / zz[row])

        # dcl (color law)
        # no color law -> dMdcl = 0

        # dMdr (recalibration)
        assert Jr is not None
        # Jr = self.recal_func.jacobian
        Jr = Jr.tocoo()
        # J_i.append(tds.spec_data.row[Jr.row])
        # J_j.append(pars['SpectrumRecalibration'].indexof(Jr.col))
        # J_val.append(Jr.data * pca[Jr.row]/zz[Jr.row])
        buff.append(tds.spec_data.row[Jr.row],
                    pars['SpectrumRecalibration'].indexof(Jr.col),
                    Jr.data * recal[Jr.row] * pca[Jr.row] / zz[Jr.row])

#        i = np.hstack(J_i)
#        j = np.hstack(J_j)
#        val = np.hstack(J_val)
#        idx = j >= 0
#        N = len(tds)
#        n_free_pars = len(pars.free)
#        JJ = scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])),
#                                     shape=(N, n_free_pars))
#
        J = buff.tocoo()
        J.data *= self.model.norm

        v = np.zeros(len(self.training_dataset))
        v[tds.spec_data.row] = self.model.norm * model_val
        return v, J


class CompressedSpectroPhotoEvalUnit(object):
    """Predicts the spectrophotometric data projected on the model wavelength basis

    Spectrophotometric spectra, hence, no recalibration polynomials, but a color law.
    """
    def __init__(self, model):
        """Constructor.

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.model = model
        self.training_dataset = model.training_dataset
        self.color_law = model.color_law

    def __call__(self, pars, jac=False):
        r"""Evaluate the model for a all spectra

        Parameters
        ----------
        pars : `saltworks.FitParameters`
            the model parameter vector
        jac : bool
            if True compute and return the jacobian matrix

        Returns
        -------
        val : numpy.array
            Model evaluations.
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (if jac is true).
        """
        tds = self.training_dataset
        spec_index = tds.spectrophot_db.spec_index
        sn_index = tds.spectrophot_db.sn_index
        mjd = tds.spectrophot_db.mjd
        z = tds.spectrophot_db.z
        #self.model.renorm(pars)
        norm = self.model.norm
        #norm = 1.

        # restframe phases
        restframe_phase = (mjd - pars['tmax'].full[sn_index]) / (1. + z)
        phase_basis = self.model.basis.by
        J = phase_basis.eval(restframe_phase + self.model.delta_phase)

        # TODO: we use the wavelenth field to store the spline basis number. We
        # need to change that: use the wavelength column to store the spline
        # function mean wavelength, and add a new field to stire the basis
        # function id.
        i_basis = tds.spectrophotometric_data.i_basis.astype(int)

        # predict the projection coefficients (and not the original spectral fluxes)
        n_wl, n_ph = len(self.model.basis.bx), len(self.model.basis.by)
        M0 = pars['M0'].full.reshape(n_ph, n_wl)
        MM = J.dot(M0)
        V0 = MM[tds.spectrophotometric_data.spec_index, i_basis]

        M1 = pars['M1'].full.reshape(n_ph, n_wl)
        MM = J.dot(M1)
        V1 = MM[tds.spectrophotometric_data.spec_index, i_basis]

        # model parameters
        X0 = pars['X0'].full[tds.spectrophotometric_data.sn_index]
        X1 = pars['X1'].full[tds.spectrophotometric_data.sn_index]
        col = pars['c'].full[tds.spectrophotometric_data.sn_index]

        # color law
        cl_pars = pars['CL'].full
        cl_pol, J_cl_pol = self.color_law(tds.spectrophotometric_data.wavelength,
                                          cl_pars, jac=jac)
        cl = np.power(10., 0.4 * col * cl_pol)

        # norm
        norm = self.model.norm

        zz = 1. + tds.spectrophotometric_data.z
        #model_val = norm * X0 * (V0 + X1 * V1) * cl / zz
        model_val = X0 * (V0 + X1 * V1) * cl / zz

        if not jac:
            v = np.zeros(len(self.training_dataset))
            v[tds.spectrophotometric_data.row] = self.model.norm * model_val
            return v

        J_i, J_j, J_val = [], [], []

        # now, the derivatives
        # dMdX0
        J_i.append(tds.spectrophotometric_data.row)
        J_j.append(pars['X0'].indexof(tds.spectrophotometric_data.sn_index))
        J_val.append(norm * (V0 + X1 * V1) * cl / zz)

        # dMdX1
        J_i.append(tds.spectrophotometric_data.row)
        J_j.append(pars['X1'].indexof(tds.spectrophotometric_data.sn_index))
        J_val.append(norm * X0 * V1 * cl / zz)

        # dMdcol
        J_i.append(tds.spectrophotometric_data.row)
        J_j.append(pars['c'].indexof(tds.spectrophotometric_data.sn_index))
        J_val.append(norm * model_val * 0.4 * np.log(10.) * cl_pol)

        # dMdtmax
        dJ = phase_basis.deriv(restframe_phase + self.model.delta_phase)
        dV0 = dJ.dot(M0)[tds.spectrophotometric_data.spec_index, i_basis]
        dV1 = dJ.dot(M1)[tds.spectrophotometric_data.spec_index, i_basis]
        J_i.append(tds.spectrophotometric_data.row)
        J_j.append(pars['tmax'].indexof(tds.spectrophotometric_data.sn_index))
        J_val.append(-norm * X0 * (dV0 + X1 * dV1) * cl / zz**2)

        del dJ

        # dMdM0
        j = np.arange(n_ph)
        row = tds.spectrophotometric_data.row.repeat(n_ph)
        tds_row = np.arange(len(tds.spectrophotometric_data)).repeat(n_ph)
        spec_index = tds.spectrophotometric_data.spec_index.repeat(n_ph)
        i_basis = tds.spectrophotometric_data.i_basis.repeat(n_ph)
        l = np.tile(np.arange(n_ph), len(tds.spectrophotometric_data))
        col = l * n_wl + i_basis # or maybe the other way around
        val = np.array(J.tocsr()[spec_index, l]).squeeze()

        J_i.append(row)
        J_j.append(pars['M0'].indexof(col))
        J_val.append(norm * X0[tds_row] * val * cl[tds_row] / zz[tds_row])

        # J_i.append(tds.spectrophotometric_data.row[J.row])
        # J_j.append(pars['M0'].indexof(J.col))
        # J_val.append(J.data * cl[J.row] / zz[J.row])

        # dMdM1
        J_i.append(row)
        J_j.append(pars['M1'].indexof(col))
        J_val.append(norm * X0[tds_row] * X1[tds_row] * val * cl[tds_row] / zz[tds_row])

        # J_i.append(tds.spectrophotometric_data.row[J.row])
        # J_j.append(pars['M1'].indexof(J.col))
        # # J.row ? really ?
        # J_val.append(X1[J.row] * J.data * cl[J.row] / zz[J.row])

        # dMdcl
        col = pars['c'].full[tds.spectrophotometric_data.sn_index]
        JJ = scipy.sparse.coo_matrix(J_cl_pol)
        J_i.append(tds.spectrophotometric_data.row[JJ.row])
        J_j.append(pars['CL'].indexof(JJ.col))
        J_val.append(norm * model_val[JJ.row] * 0.4 * np.log(10) * col[JJ.row] * JJ.data)

        i = np.hstack(J_i)
        j = np.hstack(J_j)
        val = np.hstack(J_val)
        idx = j >= 0
        N = len(tds)
        n_free_pars = len(pars.free)
        JJ = scipy.sparse.coo_matrix((val[idx], (i[idx], j[idx])),
                                     shape=(N, n_free_pars))
        #JJ.data *= self.model.norm

        v = np.zeros(len(self.training_dataset))
        v[tds.spectrophotometric_data.row] = self.model.norm * model_val
        
        return v, JJ


        # jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        # jacobian_i.append(self.data.row)
        # jacobian_j.append(self.model.pars['X0'].indexof(self.sn_index))
        # # jacobian_val.append((component_0 + x1*component_1) * color_law * recal / zz)
        # jacobian_val.append(norm * pca * color_law / zz)

        # # dMdX1
        # # jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        # jacobian_i.append(self.data.row)
        # jacobian_j.append(self.model.pars['X1'].indexof(self.sn_index))
        # # jacobian_val.append(x0 * component_1 * color_law * recal / zz)
        # jacobian_val.append(x0 * norm * component_1 * color_law / zz)

        # # dMdc
        # jacobian_i.append(self.data.row)
        # jacobian_j.append(self.model.pars['c'].indexof(self.sn_index))
        # jacobian_val.append(model * 0.4 * np.log(10.) * polynome_color_law)

        # # dMdtmax
        # # we can gain a little here, by not evaluating the gradient along the wavelength (ddlambda)
        # _, deval_phase = self.model.basis.gradient(self.restframe_wl, restframe_phases + self.model.delta_phase)
        # deval_phase = deval_phase.tocsr()
        # jacobian_i.append(self.data.row)
        # jacobian_j.append(self.model.pars['tmax'].indexof(self.sn_index))
        # #jacobian_val.append(-1. *x0* (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law * recal / zz**2)
        # jacobian_val.append(-1. * x0 * norm * (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law / zz**2)

        # # dmdtheta_0
        # jacobian_i.append(self.data.row[jacobian.row])
        # jacobian_j.append(self.model.pars['M0'].indexof(jacobian.col))
        # #jacobian_val.append(x0[jacobian.row] *jacobian.data * color_law[jacobian.row] * recal[jacobian.row] / zz[jacobian.row])
        # jacobian_val.append(x0[jacobian.row] * norm * jacobian.data * color_law[jacobian.row] / zz[jacobian.row])

        # # dmdtheta_1
        # jacobian_i.append(self.data.row[jacobian.row])
        # jacobian_j.append(self.model.pars['M1'].indexof(jacobian.col))
        # #jacobian_val.append(x0[jacobian.row]*x1[jacobian.row] * jacobian.data * color_law[jacobian.row] *
        # #                    recal[jacobian.row] / zz[jacobian.row])
        # jacobian_val.append(x0[jacobian.row] * x1[jacobian.row] * norm *  jacobian.data * color_law[jacobian.row]/ zz[jacobian.row])

        # # dMdcl (color law)
        # jacobian_color_law = scipy.sparse.coo_matrix(jacobian_color_law)
        # jacobian_i.append(self.data.row[jacobian_color_law.row])
        # jacobian_j.append(self.model.pars['CL'].indexof(jacobian_color_law.col))
        # jacobian_val.append(c[jacobian_color_law.row] * 0.4 * np.log(10.) * jacobian_color_law.data *
        #                     model[jacobian_color_law.row])

        # if self.model.disable_cache:
        #     self.model.clear_cache()
        # self.model.timing.append(time.perf_counter()-t0)


        # return model


class SpectroPhotoEvalUnit(object):
    """Evaluation of calibrated spectra
    """
    def __init__(self, tds, model):
        """Constructor.

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.training_dataset = tds
        self.data = tds.spectrophotometric_data
        self.model = model

        self.spec_index = self.data.spec_index
        self.sn_index = self.data.sn_index
        self.z = self.data.z

        # Look at that later
        self.color_law = model.color_law
        #        self.pars = model.pars
        self.basis = model.basis

        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)

        # restframe wavelengths
        self.restframe_wl = self.data.wavelength/(1.+self.z)
        #        self.Jl = model.basis.bx.eval(self.restframe_wl).tocsr()

        # and we can connect directly to the global parameters
        #        self.M0 = model.pars['M0'].full
        #        self.M1 = model.pars['M1'].full
        #        self.cl_pars = model.pars['CL'].full

    def __call__(self, pars, jac=False, debug_mode=False):
        r"""Evaluate the model for a all spectra

        Parameters
        ----------
        jac : bool
            if True compute and return the jacobian matrix
        debug_mode : bool
            if true, just return the model components.

        Returns
        -------
        val : numpy.array
            Model evaluations.
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (if jac is true).
        """

        t0 = time.perf_counter()

        # sn-related parameters
        x0, x1 = pars['X0'].full[self.sn_index], pars['X1'].full[self.sn_index]
        c, tmax = pars['c'].full[self.sn_index],  pars['tmax'].full[self.sn_index]
        # c, tmax = pars['c'].full[self.sn_index],  pars['tmax'].full[self.training_dataset.spectrophotometric_data.isn]

        # we need to re-evaluate the basis on the phases, since tmax changes
        restframe_phases = (self.data.mjd-tmax)/(1.+self.z)
        jacobian = self.basis.eval(self.restframe_wl, restframe_phases + self.model.delta_phase).tocsr()

        # norm
        norm = self.model.norm

        # model components
        component_0 = jacobian.dot(self.M0)
        component_1 = jacobian.dot(self.M1)

        polynome_color_law, jacobian_color_law = self.color_law(self.restframe_wl, self.cl_pars, jac=jac)
        color_law = np.power(10., 0.4*c*polynome_color_law)
        zz = 1. + self.z

        pca = (component_0 + x1 * component_1)
        model = x0 * norm * pca * color_law / zz

        if debug_mode:
            return component_0, component_1, color_law, recal, model

        self.model.val.append(model)
        if not jac:
            self.model.timing.append(time.perf_counter()-t0)
            return model

        jacobian = jacobian.tocoo()
        # jacobian_spec_rec = jacobian_spec_rec.tocoo()

        jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        jacobian_i.append(self.data.row)
        jacobian_j.append(pars['X0'].indexof(self.sn_index))
        # jacobian_val.append((component_0 + x1*component_1) * color_law * recal / zz)
        jacobian_val.append(norm * pca * color_law / zz)

        # dMdX1
        # jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        jacobian_i.append(self.data.row)
        jacobian_j.append(pars['X1'].indexof(self.sn_index))
        # jacobian_val.append(x0 * component_1 * color_law * recal / zz)
        jacobian_val.append(x0 * norm * component_1 * color_law / zz)

        # dMdc
        jacobian_i.append(self.data.row)
        jacobian_j.append(pars['c'].indexof(self.sn_index))
        jacobian_val.append(model * 0.4 * np.log(10.) * polynome_color_law)

        # dMdtmax
        # we can gain a little here, by not evaluating the gradient along the wavelength (ddlambda)
        _, deval_phase = self.model.basis.gradient(self.restframe_wl, restframe_phases + self.model.delta_phase)
        deval_phase = deval_phase.tocsr()
        jacobian_i.append(self.data.row)
        jacobian_j.append(pars['tmax'].indexof(self.sn_index))
        #jacobian_val.append(-1. *x0* (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law * recal / zz**2)
        jacobian_val.append(-1. * x0 * norm * (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law / zz**2)

        # dmdtheta_0
        jacobian_i.append(self.data.row[jacobian.row])
        jacobian_j.append(pars['M0'].indexof(jacobian.col))
        #jacobian_val.append(x0[jacobian.row] *jacobian.data * color_law[jacobian.row] * recal[jacobian.row] / zz[jacobian.row])
        jacobian_val.append(x0[jacobian.row] * norm * jacobian.data * color_law[jacobian.row] / zz[jacobian.row])

        # dmdtheta_1
        jacobian_i.append(self.data.row[jacobian.row])
        jacobian_j.append(pars['M1'].indexof(jacobian.col))
        #jacobian_val.append(x0[jacobian.row]*x1[jacobian.row] * jacobian.data * color_law[jacobian.row] *
        #                    recal[jacobian.row] / zz[jacobian.row])
        jacobian_val.append(x0[jacobian.row] * x1[jacobian.row] * norm *  jacobian.data * color_law[jacobian.row]/ zz[jacobian.row])

        # dMdcl (color law)
        jacobian_color_law = scipy.sparse.coo_matrix(jacobian_color_law)
        jacobian_i.append(self.data.row[jacobian_color_law.row])
        jacobian_j.append(pars['CL'].indexof(jacobian_color_law.col))
        jacobian_val.append(c[jacobian_color_law.row] * 0.4 * np.log(10.) * jacobian_color_law.data *
                            model[jacobian_color_law.row])

        if self.model.disable_cache:
            self.model.clear_cache()
        self.model.timing.append(time.perf_counter()-t0)


        return model


class SpectrumRecalibrationPolynomials:
    r"""A utility class to manage the recalibration polynomials

    The photometric calibration of spectra is generally affected by significant
    error modes at large wavelength scales. It is imperative to remove these
    error modes, while preserving the information provided by the spectra at
    small scales (spectral features)

    This is achieved during training by multiplying the spectral predictions of
    the model by a recalibration polynomial specific to each spectrum, function
    of the observation wavelength :math:`\lambda_o`, and of order :math:`N_s =
    3`, common to all spectra:

    .. math::
        s(\lambda_{rec}) = \sum_i^{N_s} s_i \lambda_{rec}^{N_s - i} %quad \mbox{and}
        \quad \lambda_{0} = \frac{\lambda - 5000}{9000 - 2700}

    with:

    .. math::
        \lambda_{rec} = \frac{2 (\lambda_o - \lambda_{max})}{\lambda_{max} - \lambda_{min}+1

    """
    def __init__(self, model, deg=3):
        """Constructor

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.model.salt
            Model.
        pol_degrees : int or indexable structure
            Polynomial degree for each spectrum.
        """

        self.tds = model.training_dataset
        self.model = model
        self.deg = deg
        self.jacobian = None

        self.n_meas = len(self.tds.spec_data)
        n_spectra = len(self.tds.spec_db)
        self.n_pars = n_spectra * (self.deg + 1)
        self.jacobian = self.build_jacobian_matrix()

    def get_struct(self):
        """
        """
        d = [('SpectrumRecalibration', self.n_pars)]
        return d

    def init_pars(self, pars):
        """initialize the model parameters

        return a parameter vector initialized such that the recalibration
        polynomials are evaluated to 1 for each spectra.

        Returns
        -------
        initialized parameters : array_like of floats

        """
        # p = np.zeros(self.n)
        # o = np.cumsum(self.pol_degrees+1) - 1
        # p[o] = 1.
        # return p
        i = np.arange(self.deg, self.n_pars, self.deg+1)
        pars['SpectrumRecalibration'].full[:] = 0.
        pars['SpectrumRecalibration'].full[i] = 1.

    def build_jacobian_matrix(self):
        """build the matrix of derivatives

        .. note:: this implementation is slightly different from what Guy had
          originally produced. Same degree for all spectra, same wavelength
          range for all spectra. If need, we may re-introduce some more
          complexity in the the future. Hopefully, we won't have to.

        """
        i, j, v = [], [], []
        spec_data = self.tds.spec_data

        wl = spec_data.wavelength
        wl_min, wl_max = wl.min(), wl.max()
        a = 2. / (wl_max - wl_min)
        b = 1. -2. * wl_max / (wl_max-wl_min)
        rwl = a * wl + b

        for spec_index in self.tds.spec_db.spec_index:
            idx = self.tds.spec_data.spec_index == spec_index
            J = scipy.sparse.coo_matrix(np.vander(rwl[idx], self.deg+1))
            nz = np.nonzero(idx)[0]
            i.append(J.row + nz.min())
            offset = (self.deg+1) * spec_index
            j.append(J.col + offset)
            v.append(J.data)

        i = np.hstack(i)
        j = np.hstack(j)
        v = np.hstack(v)
        return scipy.sparse.coo_matrix((v, (i,j)),
                                       shape=(self.n_meas, self.n_pars)).tocsr()

    def __call__(self, pars, jac=False):
        """Evaluate the recalibration polynomials

            Parameters
            ----------
            jac : bool
                If the jacobian is needed.

            Returns
            -------
            v : numpy.array
                Recalibration parameter evaluated.
            self.jacobian : None or scipy.sparse.coo_matrix
                Jacobian matrix of the recalibration polynomial functions.
            """
        p_full = pars['SpectrumRecalibration'].full
        v = self.jacobian.dot(p_full)
        if not jac:
            return v, None
        return v, self.jacobian



"""

TODO: Put the iterative training steps.
"""
import numpy as np
import os
import pylab as pl
#import corner
import pandas

from nacl.models.salt import SALT2Like
from nacl import models
from .lib import bspline
from . import minimize
# from .minimizers import Minimizer, ModelPulls
#from .models.constraints import Constraints2D
#from .models.regularizations import SplineRegul2D
#from .models.variancemodels import SimpleVarianceModel
#from nacl.minimizers import Minimizer
#from nacl.models.regularizations import NaClSplineRegularization
#from .models import constraints
#from nacl.models.constraints import SALT2LikeConstraints, ab_flux_at_10Mpc

#from .models.salt2.constraints import Constraints2D
#from .models.salt2.regularizations import SplineRegul2D
#from .models.salt2.variancemodels import SimpleVarianceModel
from nacl.minimizers import Minimizer
from nacl import minimize
#from nacl.models.salt2.regularizations import NaClSplineRegularization
from .models import constraints
#from nacl.models.salt2.constraints import SALT2LikeConstraints, ab_flux_at_10Mpc


from .plotting import plots

from nacl.models import salt2
from nacl.loglikelihood import LogLikelihood

try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt



def scramble_init(pars):
    """
    """
    sz = len(pars['X0'].free)
    pars['X0'].free *= np.random.normal(size=sz, loc=1., scale=0.1)
    pars['X1'].free += np.random.normal(size=sz, scale=1.)
    pars['col'].free += np.random.normal(size=sz, scale=0.2)
    pars['tmax'].free += np.random.normal(size=sz, scale=1.)


class TrainSALT2Like:

    def __init__(self, tds, linear_constraints=True, mu_reg=1, reg_order=1, mu_cons=1.E6, Mb=-19.5):
        """
        """
        self.tds = tds
        self.linear_constraints = linear_constraints
        self.mu_reg = mu_reg
        self.reg_order = reg_order
        self.mu_cons = mu_cons
        self.Mb = Mb
        self.log = []
        #self.model = models.get_model(tds)
        self.model = salt2.get_model(tds)
        self.cons = []
        self.reg = []
        self.var = []

    def fit(self, block_names, pars=None,
            accept=10., reject=10., max_iter=100,
            lambda_init=1.E3, diag_charge='levenberg',
            dchi2_stop=0.01, fit_error_model=False,
            spec_weight_scale=1.):
        """perform one likelihood minimization
        """
        # starting point (if available)
        if pars is None:
            pars = self.model.init_pars()

        # model : release only the requested parameters
        pars.fix()
        for block_name in block_names:
            pars[block_name].release()

        # instantiate the regularization scheme
        # need to re-instantiate it, since M0 M1 may be either free or fixed
        #reg = models.get_regularization(self.model, mu=self.mu_reg, order=self.reg_order, check=True)
        reg = salt2.get_regularization_prior(self.model, pars=pars, mu=self.mu_reg, order=self.reg_order, check=True)
        self.reg.append(reg)
        
        # instantiate the constraints
        # to to re-instantiate it, since we may or may not fit the model
        #cons = models.get_constraints(self.model, linear=self.linear_constraints,
        #                              mu=self.mu_cons, Mb=self.Mb,
        #                              check=True)
        cons = salt2.get_constraint_prior(self.model, linear=self.linear_constraints,
                                      mu=self.mu_cons, Mb=self.Mb,
                                      check=True)
        self.cons.append(cons)

        # instantiate a LogLikelihood, with or without an error model
        error_snake = calib_scatter = color_scatter = None
        if fit_error_model:
            error_snake = salt2.get_simple_snake_error(self.model)
            self.var.append(error_snake)

        #-----FIRST FIT WITHOUT ERROR_SNAKE--------
        ll = LogLikelihood(self.model, reg=[reg], cons=[cons],
                                   force_default_spgemm=False,
                                   variance_model=None)
                                   # the other components will be added here
                                   
        ll.pars.fix()
        for block_name in block_names:
            ll.pars[block_name].release()
        # instantiate a minimizer
        minz = minimize.Minimizer(ll)

        # minimize - as of today, we use the Levenberg-Marquardt algorithm
        res = minz.minimize_lm(pars.free,
                               dchi2_stop=dchi2_stop, max_iter=max_iter,
                               max_attempts=60, accept=accept, reject=reject,
                               lamb=lambda_init,
                               diag_charge=diag_charge,
                               )
        minz.plot()
        #-------SECOND FIT WITH ERROR SNAKE---------
        ll = LogLikelihood(self.model, variance_model=error_snake, force_default_spgemm=False)
        
        for block_name in ['X0', 'X1', 'col', 'M0', 'M1', 'CL', 'tmax']:
            ll.pars[block_name].full[:] = res['pars'][block_name].full[:]
        ll.pars.fix()
        for block_name in block_names:
            ll.pars[block_name].release()  
        reg = salt2.get_regularization_prior(self.model, mu=self.mu_reg, order=self.reg_order, check=True, pars=ll.pars)
        
        ll = LogLikelihood(self.model, reg=[reg], cons=[cons], 
                                   variance_model=error_snake,
                                   force_default_spgemm=False)
                                   #spec_weight_scale=spec_weight_scale,
                                   # the other components will be added here
                                   
        for block_name in ['X0', 'X1', 'col', 'M0', 'M1', 'CL', 'tmax']:
            ll.pars[block_name].full[:] = res['pars'][block_name].full[:]
        # instantiate a minimizer
        minz = minimize.Minimizer(ll)

        # minimize - as of today, we use the Levenberg-Marquardt algorithm
        res = minz.minimize_lm(ll.pars.free,
                               dchi2_stop=dchi2_stop, max_iter=max_iter,
                               max_attempts=60, accept=accept, reject=reject,
                               lamb=lambda_init,
                               diag_charge=diag_charge,
                               )
        
        minz.plot()
        log = minz.get_log()
        res['log'] = log
        res['minz'] = minz # remove this in the future
        self.log.append(log)

        return res

    def fit_by_blocks(self, block_names, pars=None,
            accept=10., reject=10., max_iter=100,
            lambda_init=1.E3, diag_charge='levenberg',
            dchi2_stop=0.01, fit_error_model=False,
            spec_weight_scale=1.):
        """perform 2 likelihood minimizations
        """
        if pars is None:
            pars = self.model.init_pars()

        pars.fix()
        for block_name in block_names:
            pars[block_name].release()

        reg = salt2.get_regularization_prior(self.model, pars=pars, mu=self.mu_reg, order=self.reg_order, check=True)
        self.reg.append(reg)

        cons = salt2.get_constraint_prior(self.model, linear=self.linear_constraints,
                                      mu=0., Mb=self.Mb,
                                      check=True)
        self.cons.append(cons)

        error_snake = calib_scatter = color_scatter = None
        

        #-----FIRST FIT WITHOUT ERROR_SNAKE--------
        ll = LogLikelihood(self.model, reg=[reg], cons=[cons],
                                   force_default_spgemm=False,
                                   variance_model=None)
                                   # the other components will be added here
                                   
        ll.pars.fix()
        for block_name in block_names:
            ll.pars[block_name].release()
        # instantiate a minimizer
        minz = minimize.Minimizer(ll)

        # minimize - as of today, we use the Levenberg-Marquardt algorithm
        res = minz.minimize_lm(pars.free,
                               dchi2_stop=dchi2_stop, max_iter=max_iter,
                               max_attempts=60, accept=accept, reject=reject,
                               lamb=lambda_init,
                               diag_charge=diag_charge,
                               )
        minz.plot()
        #-------SECOND FIT WITH ALL RELEASED---------
        ll.pars.release()
          
        reg = salt2.get_regularization_prior(self.model, mu=self.mu_reg, order=self.reg_order, check=True, pars=ll.pars)
        cons = salt2.get_constraint_prior(self.model, linear=self.linear_constraints,
                                      mu=self.mu_cons, Mb=self.Mb,
                                      check=True)

        ll = LogLikelihood(self.model, reg=[reg], cons=[cons], 
                                   variance_model=error_snake,
                                   force_default_spgemm=False)
                                   
        for block_name in ['X0', 'X1', 'col', 'M0', 'M1', 'CL', 'tmax']:
            ll.pars[block_name].full[:] = res['pars'][block_name].full[:]
        # instantiate a minimizer
        ll.pars.release()
        minz = minimize.Minimizer(ll)

        # minimize - as of today, we use the Levenberg-Marquardt algorithm
        res = minz.minimize_lm(ll.pars.free,
                               dchi2_stop=dchi2_stop, max_iter=max_iter,
                               max_attempts=60, accept=accept, reject=reject,
                               lamb=lambda_init,
                               diag_charge=diag_charge,
                               )
        minz.plot()
        #-------THIRD FIT WITH ALL RELEASED AND ERROR MODEL---------   
        if fit_error_model:
            error_snake = salt2.get_simple_snake_error(self.model)
            self.var.append(error_snake)               
        if error_snake is not None:
            ll = LogLikelihood(self.model, variance_model=error_snake, force_default_spgemm=False)
            
            for block_name in ['X0', 'X1', 'col', 'M0', 'M1', 'CL', 'tmax']:
                ll.pars[block_name].full[:] = res['pars'][block_name].full[:]
            #ll.pars.fix()
            #for block_name in block_names:
            #    ll.pars[block_name].release()
              
            reg = salt2.get_regularization_prior(self.model, mu=self.mu_reg, order=self.reg_order, check=True, pars=ll.pars)
            cons = salt2.get_constraint_prior(self.model, linear=self.linear_constraints,
                                          mu=self.mu_cons, Mb=self.Mb,
                                          check=True)

            ll = LogLikelihood(self.model, reg=[reg], cons=[cons], 
                                       variance_model=error_snake,
                                       force_default_spgemm=False)
                                       
            for block_name in ['X0', 'X1', 'col', 'M0', 'M1', 'CL', 'tmax']:
                ll.pars[block_name].full[:] = res['pars'][block_name].full[:]
            # instantiate a minimizer
            minz = minimize.Minimizer(ll)

            # minimize - as of today, we use the Levenberg-Marquardt algorithm
            res = minz.minimize_lm(ll.pars.free,
                                   dchi2_stop=dchi2_stop, max_iter=max_iter,
                                   max_attempts=60, accept=accept, reject=reject,
                                   lamb=lambda_init,
                                   diag_charge=diag_charge,
                                   )
            minz.plot()
        log = minz.get_log()
        res['log'] = log
        res['minz'] = minz
        self.log.append(log)

        return res

    def train(self, lambda_init_first_fit=1.E3, max_iter_first_fit=15,
              spec_weight_scale=1.,
              diag_charge='levenberg'):
        """train the model and error model

        this method implements the standard training scheme, described in GA's thesis:
          - fit the model (no error model)
          - fit the error model (fixed model and X pars)
          - final fit with all parameters released
        """
        self.log = []

        # first fit, no error model
        # limited number of iterations
        # self.model.pars['X1'].full -= self.model.pars['X1'].full.mean()
        # self.model.pars['X1'].full /= self.model.pars['X1'].full.std()
        r = self.fit(['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL',
#                      'SpectrumRecalibration'
                      ],
                     accept=10., reject=10., max_iter=max_iter_first_fit,
                     spec_weight_scale=spec_weight_scale,
                     lambda_init=lambda_init_first_fit,
                     dchi2_stop=0.01, fit_error_model=False,
                     diag_charge=diag_charge)

        # second fit: error model only
        # TODO: how do we specify the error model parameters ?
        self.model.pars['sigma_snake'].full[:] = 0.05
        r = self.fit(['sigma_snake', 'eta_calib'],
                     accept=10., reject=10., max_iter=100,
                     spec_weight_scale=spec_weight_scale,
                     lambda_init=1.E-3,
                     dchi2_stop=0.01, fit_error_model=True,
                     diag_charge=diag_charge)

        # last fit: all pars released
        # TODO: abbrev for "all pars"
        # self.mu_reg = 1.E-3
        self.mu_cons = 1.E12
        r = self.fit(['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL',
#                      'SpectrumRecalibration',
                      'sigma_snake', 'eta_calib'],
                     accept=10., reject=2., max_iter=100,
                     spec_weight_scale=spec_weight_scale,
                     lambda_init=10.,
                     dchi2_stop=0.01, fit_error_model=True,
                     diag_charge=diag_charge)

        return r

    def plot(self, r=None):
        """
        """
        pass

    def plot_residuals(self):
        """
        """
        pass

    def plot_pars_vs_truth(self):
        """
        """
        pass

    def plot_model_vs_truth(self):
        """
        """
        pass


def plot_pars_vs_truth(model, p_truth):
    """
    """
    p_rec = model.pars
    dp_x0   = (p_rec['X0'].full - p_truth['X0'].full) / p_truth['X0'].full
    dp_x1   = p_rec['X1'].full-p_truth['X1'].full
    dp_col  = p_rec['col'].full-p_truth['col'].full
    dp_tmax = p_rec['tmax'].full-p_truth['tmax'].full
    samples = pandas.DataFrame({'X0': dp_x0, 'X1': dp_x1, 'col': dp_col, 'tmax': dp_tmax})
    fig = corner.corner(samples, bins=20, plot_contours=False,
                        # truths=[0., 0., 0., 0.],
                        # truth_color='gray',
                        hist_bin_factor=1.,
                        labels=['$X_0$', '$X_1$', '$col$', '$t_{max}$'],
                        show_titles=True)
    fig.suptitle('fitted - truth [SN pars]')
    return fig

def plot_sn_pars_vs_z(model, p_truth):
    """
    """
    p_rec = model.pars
    tds = model.training_dataset

    fig, axes = pl.subplots(nrows=2, ncols=2, figsize=(8,8), sharex=True)
    dp = (p_rec['X0'].full - p_truth['X0'].full) / p_truth['X0'].full
    axes[0,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index],'k.')
    axes[0,0].set_title('X0')
    axes[0,0].set_ylabel('$X_0^{rec} - X_0^{true}$')

    dp = p_rec['X1'].full-p_truth['X1'].full
    axes[0,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
    axes[0,1].set_title('X1')
    axes[0,1].set_ylabel('$X_1^{rec} - X_1^{true}$')

    dp = p_rec['col'].full-p_truth['col'].full
    axes[1,0].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
    axes[1,0].set_title('$col$')
    axes[1,0].set_ylabel('$col^{rec} - col^{true}$')

    dp = p_rec['tmax'].full-p_truth['tmax'].full
    axes[1,1].plot(tds.sn_data.z, dp[tds.sn_data.sn_index], 'k.')
    axes[1,1].set_title('$t_{max}$')
    axes[1,1].set_ylabel('$t_{max}^{rec} - t_{max}^{true}$')

def plot_minimization_logs(logs, vs_time=False):
    """
    """
    fig, axes = pl.subplots(nrows=6, ncols=1, sharex=True, figsize=(14,10))
    dstep = 0
    t0 = None

    for i,log in enumerate(logs):
        ii = (i+1) * int(256/len(logs))
        if t0 is None:
            t0 = log.time.min()
        col = pl.cm.jet(ii)

        if vs_time:
            x = log.time - t0
        else:
            x = log.step + dstep

        axes[0].semilogy(x, log.chi2, ls=':', marker='.', color=col)
        axes[0].set_ylabel('$\chi^2$')

        if i == 0:
            axes[1].semilogy(x[-3:], log.chi2[-3:], ls=':', marker='.', color=col)
        else:
            axes[1].semilogy(x, log.chi2, ls=':', marker='.', color=col)
        axes[1].set_ylabel('$\chi^2$ (zoom)')

        axes[2].semilogy(x, log.dchi2, ls=':', marker='.', color=col)
        axes[2].set_ylabel('$\Delta\chi^2$')

        if log.log_det_v.any():
            axes[3].plot(x, log.log_det_v, ls=':', marker='.', color=col)
        axes[3].set_ylabel('$|\mathbf{V}|$')


        axes[4].semilogy(x, log.cons, ls=':', marker='.', color=col)
        axes[4].set_ylabel('$\chi^2_{cons}$')

        axes[5].plot(x, log.reg, ls=':', marker='.', color=col)
        axes[5].set_ylabel('$\chi^2_{reg}$')
        axes[5].set_xlabel('step')

        dstep += log.step.max()

    axes[0].set_title('training fit')
    pl.subplots_adjust(hspace=0.0025)


def plot_surface(model, surface_name='M0', p_other=None):
    """
    """
    if p_other is not None:
        fig, axes = pl.subplots(nrows=3, ncols=1, figsize=(8,16), sharex=True, sharey=True)
        ax = axes[0]
    else:
        fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(8,8))
        ax = axes

    n_phase = len(model.basis.by)
    n_wl = len(model.basis.bx)
    im = model.pars[surface_name].full.reshape((n_phase, n_wl))
    ax.imshow(im)
    if p_other is not None:
        im_other = p_other[surface_name].full.reshape(n_phase, n_wl)
        axes[1].imshow(im_other)
        axes[2].imshow(im - im_other)

def plot_colorlaw(model):
    """
    """
    pass


def fit(model, block_names, init_pars=None,
        reg_order=0, mu_reg=1.E-6, mu_cons=1.E6,
        accept=10., reject=10.,
        Mb=-19.5, n_iter=100):
    """
    """
    if init_pars is not None:
        model.pars.free = init_pars

    model.pars.fix()
    for block_name in block_names:
        model.pars[block_name].release()

    # we may need some regularization
    reg = []
    to_regularize = []
    for block_name in ['M0', 'M1']:
        if block_name in block_names:
            to_regularize.append(block_name)
    if len(to_regularize) > 0:
        reg = NaClSplineRegularization(model, to_regularize=to_regularize,
                                       order=reg_order, mu=mu_reg)
        reg = [reg]

    # we may need some contraints
    cons = []
    active = {}
    for cons_name in ['M0', 'dM0', 'M1', 'dM1', 'col', 'X1', 'X1_var']:
        if 'M0' in block_names and 'X0' in block_names:
            active['M0'] = ab_flux_at_10Mpc(Mb)
        if 'M1' in block_names:
            active['M1'] = 0.
        if 'M0' in block_names and 'tmax' in block_names:
            active['dM0'] = 0.
        if 'M1' in block_names and 'tmax' in block_names:
            active['dM1'] = 0.
        if 'CL' in block_names and 'col' in block_names:
            active['col'] = 0.
        if 'X1' in block_names and 'M1' in block_names:
            active['X1'] = 0.
            active['X1_var'] = 1.
    if len(active) > 0:
        cons = [SALT2LikeConstraints(model, active=active, mu=mu_cons)]

    # if there are constraints, make sure that they are fullfilled
    # at first order
    model.pars['X1'].full -= model.pars['X1'].full.mean()
    model.pars['X1'].full /= model.pars['X1'].full.std()
    model.pars['col'].full -= model.pars['col'].full.mean()

    # fit
    wres = minimize.WeightedResiduals(model)
    chi2 = minimize.LogLikelihood(wres, reg=reg, cons=cons)
    minz = minimize.Minimizer(chi2, max_iter=n_iter)    #n_inter should be max_iter?

    # start the fit
    solution = minz.minimize(model.pars.free)

    ret = {}
    ret['solution'] = solution
    ret['model'] = model
    ret['cons'] = cons
    ret['reg'] = reg
    ret['chi2'] = chi2
    ret['pars'] = model.pars.copy()

    return ret


def train_salt2like(model, lcfit=False, mu_reg=1., mu_cons=1.E6):
    """
    """
    ret_info = {}

    # fit #0 [optional], fixed model, only the LC parameters allowed
    if lcfit:
        r = fit(model, ['X0', 'X1', 'col', 'tmax'])
        ret_info['fit0'] = r

    # fit #1: LC + model, no error model
    r = fit(model, ['X0', 'X1', 'col', 'tmax',
                    'M0', 'M1', 'CL'],
            accept=10., reject=10,
            mu_reg=mu_reg, mu_cons=mu_cons)
    ret_info['fit1'] = r

    # fit #2: error model only
    r = fit(model, ['sigma_snake'],
            accept=10., reject=10.,
            mu_reg=mu_, mu_cons=mu_cons)
    ret_info['fit2'] = r

    # fit #3: all pars released
    r = fit(model, ['X0', 'X1', 'col', 'tmax',
                    'M0', 'M1', 'CL',
                    'sigma_snake'],
            accept=10., reject=2.,
            mu_reg=mu_reg, mu_cons=mu_cons)
    ret_info['fit3'] = r



class FitModel2D(object):
    """
    Fitting procedure class.
    From a simulation, a model, error models, and parameters,
    can fit the model (given a minimizer) on data.
    If parameters are fixed, regularization and constraints are adjusted.

    All the parameters of the fit, amplitude of constraints, regularization can be specified.

    Attributes
    ----------
    training_dataset : nacl.dataset.TrainingDataset
        Training dataset.
    model : nacl.salt
        Model.
    gamma_init : numpy.array ou float
        Parameters used to initialize the error snake.
    basis_knot : 2-list
        Number of wavelength and photometric knot.
    init_from_salt2_file : str
        If the model is to be initialized with the original
        SALT2-4 surfaces and bases, get them from this file.
    norm_band : str
        Use this band to normalize the model.
    degree : int
        Degree of the spectrum recalibration polynomials
    BETA : float
        Diagonal offset in hessian matrix where use of cholesky.
    queue_wsp : list
        Model queue of eval unit.
    pars_free : list
        List of free parameters in the fit.
    beta : numpy.array
        Fitted model parameter.
    beta0 : numpy.array
        Model parameter initialization.
    reg : nacl.models.regularizations
        Class of regularization.
    pars_generation : numpy.array
        Parameters used for the simulation.
    all_pars : list
        List of all the parameters of the model.
    cons : nacl.models.constraints
        Class of constraints.
    x1_var_real : float
        Amplitude of constraints to the total :math:`\chi^2`
    N_sn_4plot : int
        Index of SN data to plot.
    variance_model : nacl.models.variancemodels.SimpleVarianceModel
        Error snake model.
    gamma : numpy.array
        Fitted error snake parameter :math:`\gamma`
    gamma0 : numpy.array
        Error snake parameter :math:`\gamma` at initialisation
    kappa_init : numpy.array
        Color scatter parameters.
    color_scatter_func : object
        Color scatter function used.
    color_scatter : nacl.models.variancemodels.ColorScatter
        Color scatter model.
    mu_pq : float
        Amplitude of the linear constraints to the total :math:`\chi^2`.
    log : list
        For debugging reason, stored at each step all fit quantities.
    N_ITER : int
        Maximum number of fit iteration.
    df_stop : float
        :math:`\Delta \chi^2` after which the fit stops.
    pars_fix : list
        List a parameter fixed during the fit.
    mu_pq_var_x1 : float
        Amplitude of constraints on :math:`X_1` variance to the total :math:`\chi^2`
    mu_reg : float
        Amplitude of the regularization to the total :math:`\chi^2`
    chi2_normalization : float
        Ratio of :math:`\chi^2` of spectral data and :math:`\chi^2` of photometric data.
    minimize_func : nacl.minimizers.Minimizer
        Minimizer.
    log : list
        For debugging reason, stored at each step all fit quantities.
    df_stop : float
        :math:`\Delta \chi^2` after which the fit stops.
    pars_fix : list
        List a parameter fixed during the fit.

    """
    def __init__(self, training_dataset, model_func, normalization_band_name='SWOPE::B',
                 init_from_salt2_file='data/salt2.npz', pars_fix=[],
                 minimizer=Minimizer, degree=3, disconnect_sp=False,
                 gamma_init=0.05, kappa_init=None, x1_var_real=False,
                 mu_pq=1e10, mu_pq_var_x1=None, mu_reg=1,
                 variance_model_bins=(3,3),
                 chi2_normalization=1, color_scatter_func=None, beta=1e-10,
                 log=True, df_stop=1e-1, n_iter=500, basis_knots=[127, 20],
                 pars_generation=None):
        """
        Constructor.

        Parameters
        ----------
        training_dataset : nacl.dataset.TrainingDataset
            Training dataset.
        model_func : nacl.salt
            Model.
        gamma_init : numpy.array ou float
            Parameters used to initialize the error snake.
        basis_knots : 2-list
            Number of wavelength and photometric knot.
        init_from_salt2_file : str
            If the model is to be initialized with the original
            SALT2-4 surfaces and bases, get them from this file.
        normalization_band_name : str
            Use this band to normalize the model.
        degree : int
            Degree of the spectrum recalibration polynomials
        beta : float
            Diagonal offset in hessian matrix where use of cholesky.
        pars_generation : numpy.array
            Parameters used for the simulation.
        x1_var_real : float
            Amplitude of constraints to the total :math:`\chi^2`
        N_sn_4plot : int
            Index of SN data to plot.
        kappa_init : numpy.array
            Color scatter parameters.
        color_scatter_func : nacl.models.variancemodels.ColorScatter
        mu_pq : float
            Amplitude of the linear constraints to the total :math:`\chi^2`.
        log : list
            For debugging reason, stored at each step all fit quantities.
        n_iter : int
            Maximum number of fit iteration.
        df_stop : float
            :math:`\Delta \chi^2` after which the fit stops.
        pars_fix : list
            List a parameter fixed during the fit.
        mu_pq_var_x1 : float
            Amplitude of constraints on :math:`X_1` variance to the total :math:`\chi^2`
        mu_reg : float
            Amplitude of the regularization to the total :math:`\chi^2`
        variance_model_bins : tuple(int, int)
            Number of bins of the vraince surface
        chi2_normalization : float
            Ratio of :math:`\chi^2` of spectral data and :math:`\chi^2` of photometric data.
        minimizer : nacl.minimizers.Minimizer
            Minimizer.
        log : list
            For debugging reason, stored at each step all fit quantities.
        df_stop : float
            :math:`\Delta \chi^2` after which the fit stops.
        pars_fix : list
            List a parameter fixed during the fit.

        """
        self.training_dataset = training_dataset
        self.norm_band = normalization_band_name
        self.degree = degree
        self.BETA = beta
        self.gamma_init = gamma_init
        self.kappa_init = kappa_init
        self.color_scatter_func = color_scatter_func
        self.x1_var_real = x1_var_real

        self.pars_free, self.x0, self.g0, self.reg, self.cons = None, None, None, None, None

        self.basis_knots = basis_knots
        self.model = model_func(self.training_dataset,
                                normalization_band_name=self.norm_band,
                                spectrum_recal_degree=self.degree,
                                basis_knots=self.basis_knots)

        self.queue_wsp = self.model.queue.copy()

        self.init_model_pars(filename=init_from_salt2_file, sn_info=self.training_dataset.sn_data)

        self.variance_model = SimpleVarianceModel(self.model,
                                                  gamma_init=self.gamma_init,
                                                  bins=variance_model_bins)

        self.pars_generation = pars_generation
        self.N_sn_4plot = None
        self.all_pars = ['X0', 'tmax', 'X1', 'col', 'M0', 'M1', 'CL',
                         'SpectrumRecalibration', 'eta_calib', 'kappa_color']

        self.mu_pq = mu_pq
        self.mu_pq_var_x1 = mu_pq_var_x1
        self.mu_reg = mu_reg
        self.chi2_normalization = chi2_normalization
        self.minimize_func = minimizer

        self.log = log
        self.N_ITER = n_iter
        self.df_stop = df_stop

        self.pars_fix = pars_fix
        self.pars_release()
        self.model.pars0 = self.model.pars.full[:].copy()


    def init_model_pars(self, filename, sn_info):
        r"""
        Initiate model parameters.

        Parameters
        ----------
        filename : str
            The original
            SALT2-4 surfaces and bases, get them from this file.
        sn_info : numpy.rec.array
            SNe information (:math:`(z, X_0, X_1, c, t_{max})`)
        """
        m0, m1, cl = self.init_with_salt2(filename, self.model)

        self.model.pars['X0'][:] = sn_info.x0
        self.model.pars['X1'][:] = sn_info.x1
        self.model.pars['tmax'][:] = sn_info.tmax
        self.model.pars['col'][:] = sn_info.col

        self.model.pars['M0'][:] = m0
        self.model.pars['M1'][:] = m1
        self.model.pars['CL'][:] = cl
        self.model.norm = self.model.normalization(band_name=self.norm_band)

    @staticmethod
    def init_with_salt2(filename, model):
        """
        Initiate the model with salt2.4 files.

        Parameters
        ----------
        filename : str
            The original SALT2-4 file.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        f = np.load(filename)
        phase_grid = f['phase_grid']
        wl_grid = f['wl_grid']
        basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)

        xx = np.linspace(wl_grid[0], wl_grid[-1], basis.bx.nj)
        yy = np.linspace(phase_grid[0], phase_grid[-1], basis.by.nj)

        x, y = np.meshgrid(xx, yy)
        x, y = x.ravel(), y.ravel()
        jacobian = model.basis.eval(x, y)
        factor = cholesky_AAt(jacobian.T, beta=1e-20)

        return factor(jacobian.T * f['M0']), factor(jacobian.T * f['M1']), f['CL_pars']

    def pars_release(self):
        """
        Fixed the parameters label a pars_fix.
        """
        self.pars_free = [i for i in self.all_pars if i not in self.pars_fix]
        self.model.pars.fix()
        for pr in self.pars_free:
            self.model.pars.release(pr)

    def parameter_degradation(self, model_coef_init=0.05,
                              var_coef_init=0.1,
                              seed=None):
        """
        Degrade the parameters.
        Multiply their values by normal distribution, of mean 1 and a chosen standard deviations.
        For :math:`t_{max}` add normal distribution, of mean 0 and a chosen standard deviations.

        Parameters
        ----------
        model_coef_init : float
            Standard deviation for the model parameters : :math:`(X_0, X_1, c, t_{max})`
        var_coef_init : float
            Standard deviation for the error snake parameters : :math:`\gamma`
        seed : int
            Seed for numpy.random
        """
        if seed is not None:
            np.random.seed(seed)

        self.variance_model.pars.full[:] *= np.random.normal(1., var_coef_init,
                                                             len(self.variance_model.pars.full))
        print(f'\nVarModPars times N(1, {var_coef_init})')
        nb_sn = len(self.model.training_dataset.sne)
        for ps in self.pars_free:
            if (ps != 'eta_calib') & (ps != 'tmax'):
                mv = model_coef_init
                self.model.pars[ps].full[:] *= np.random.normal(1., mv,
                                                                len(self.model.pars[ps].full))
                if ps == 'X1':
                    self.model.pars[ps].full[:] -= self.model.pars[ps].full[:].mean()

                print(f'{ps} times N(1, {mv})')

            elif model_coef_init != 0.:
                if ps == 'tmax':
                    s_tmax = 0.05
                    self.model.pars['tmax'].full[:] += np.random.normal(0, s_tmax, nb_sn)
                    print(f'{ps} + N(0, {s_tmax})')
        print('\n')
        self.gamma0 = self.variance_model.pars.free.copy()
        self.beta0 = self.model.pars.free.copy()

    def regularization(self, sur_reg=np.array(['M0', 'M1'])):
        """
        Instantiate regularization class.

        Parameters
        ----------
        sur_reg : numpy.array
            Surfaces to regularized.
        """
        sur_id = np.in1d(sur_reg, self.pars_free)
        if sur_id.sum() != 0:
            print(f'reg : {sur_reg[sur_id]}')
            self.reg = SplineRegul2D(self.model, mu=self.mu_reg,
                                     surfaces_reg=sur_reg[sur_id])
        else:
            self.reg = None

    def constraints(self, paramaters_cons=np.array(['M0', 'M1', 'CL', 'X0', 'tmax', 'X1', 'c'])):
        """
        Instantiate regularization class.

        Parameters
        ----------
        paramaters_cons : numpy.array
            List of parameters on which a constraint is applied.
        """
        cons_id = np.in1d(paramaters_cons, self.pars_free)

        # if model is fixed no need of contains
        if np.in1d(np.array(['M0', 'M1', 'CL']), self.pars_fix).sum() != 0.:
            self.cons = None

        elif cons_id.sum() != 0:
            cc = [i for i in paramaters_cons[cons_id]]
            print(f'cons : {cc}')
            self.cons = Constraints2D(self.model, mu_pq=self.mu_pq, parameters_cons=cc,
                                      mu_pq_var_x1=self.mu_pq_var_x1,
                                      x1_var_real=self.x1_var_real)   # ,vals = [])

        else:
            self.cons = None

    def __call__(self, pars_fix=[],
                 g=None, eta_covmatrix=None, sigma_kappa=None,
                 fit_spec=True, sigma_kappa_fit=True):
        """
        Fit !

        Parameters
        ----------
        pars_fix : list
            List of fixed parameters during the fit. If the error snake should
            be fixed 'gamma' should appear here.
        g : np.array
            Error snake parameters free or fixed.
        eta_covmatrix : float or numpy.array
            Calibration uncertainty matrix.
        sigma_kappa :
            Color scatter parameter.
        fit_spec : bool
            If spectral data are considered.
        sigma_kappa_fit : bool
            If color scatter should be fixed.

        TODO : add color scatter parameters fixation in pars_fix.
        """

        self.pars_fix = np.array(pars_fix)
        self.pars_release()

        print("\nNumber of free parameters :")
        for i in self.all_pars:
            print(i, self.model.pars[i].free.shape)
        print("\n")

        self.regularization()
        self.constraints()
        self.g = g
        if fit_spec is False:
            self.model.queue = [i for i in self.model.queue if (i.data['spec_id'].mean() == -1)]

        self.f = ModelPulls(self.model,
                            variance_model=self.variance_model, fit_spec=fit_spec,
                            chi2_normalization=self.chi2_normalization)

        if 'gamma' in pars_fix:
            self.f.VarianceModel.pars.fix()

        if sigma_kappa is not None:
            self.color_scatter = self.color_scatter_func(self.model.lambda_c_red, sigma_kappa)
        else :
            self.color_scatter = None

        minimizer = self.minimize_func(self.f, constrains=self.cons,
                                       regularization=self.reg,
                                       color_scatter=self.color_scatter, ##
                                       n_iter=self.N_ITER,
                                       df_stop=self.df_stop, log=self.log, beta_cholmod=self.BETA)

        minimizer(self.model.pars.free, sigma_kappa=sigma_kappa,
                  gamma0=self.g, eta_covmatrix=eta_covmatrix, sigma_kappa_fit=sigma_kappa_fit)

        self.M = minimizer.M
        self.gradf = minimizer.gradf
        self.minim = minimizer

        self.sigma_kappa = minimizer.sigma_kappa
        if (self.sigma_kappa is not None) & sigma_kappa_fit:
            self.dsigma_kappa = minimizer.dsigma_kappa

        self.timing = {'iter': np.array(minimizer.titer).mean(),
                       'cholesky': np.array(minimizer.tcholesky).mean(),
                       'model': np.array(minimizer.tmodel).mean(),
                       'hessian': np.array(minimizer.thessian).mean(),
                       's_iter': np.array(minimizer.titer).std(),
                       's_cholesky': np.array(minimizer.tcholesky).std(),
                       's_model': np.array(minimizer.tmodel).std(),
                       's_hessian': np.array(minimizer.thessian).std()}

        self.log = minimizer.log

        self.beta = minimizer.beta
        self.gamma = minimizer.gamma

        if fit_spec is False:
            self.model.queue = self.queue_wsp

    def control_plot(self, n_fit, variance_model=None,
                     theta_init=None, save_dir=None, add_name=''):
        """
        Make control plots.
        """
        if save_dir is not None:
            save_dir = f'{save_dir}'
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)

        if variance_model is not None:
            variance_model = self.variance_model
        if self.N_sn_4plot is None:
            self.N_sn_4plot = np.random.randint(0, len(self.training_dataset.sne), 5)

        for isn in self.N_sn_4plot:
            plots.plot_lc(self.model, isn, variance_model=variance_model,
                          save_dir=save_dir, add_name=add_name)
            plots.plot_spec(self.model, isn, variance_model=variance_model,
                            save_dir=save_dir, add_name=add_name)
        # plots.binning_residuals(self.model, save_dir=save_dir, add_name=add_name)
        plots.residual_histogram(self.model, n_fit, variance_model=variance_model,
                                 data_selection=False, save_dir=save_dir)

        if variance_model:
            plots.plot_2Dsurfacemodel(variance_model, ['gamma'],

                                      vmin=0.90, vmax=1.1, ddd=False,
                                      init_unit=True, theta_init=theta_init,
                                      save_dir=save_dir, add_name=add_name)
